# -*- coding: utf-8 -*-
"""
.. Authors
     Novimir Pablant <npablant@pppl.gov>

A set of tools to save/load python dictionaries to/from hdf5 files.

hdf5 files are an attractive standard for storing scientific data as they have
a standardized format and can be read from any language or system.

This project aims to make saving/loading of python dictionaries or lists to
the hdf5 format completely seamless for standard python and numpy data types.
To achieve this we add extra information into the hdf5 file through the use
of dataset attributes so that when loaded, the original dictionary can be
fully restored.

Programming Notes
-----------------

Python specific types/objects cannot in general be directly translated to
HDF5 data types.  The use of non compatible objects/types while using h5py will
produce an error.  There are however a few exceptions that have been added
to mirHDF5 for convenience.

None
  The python None object will be saved as False, with specific
  HDF5 attributes to allow proper restortation in python.

Dict
  Python dictionaries are saved using HDF5 group and data entries.

List
  Python lists are saved using HDF5 group and data entries.  Entry
  names are automatically generated, and an HDF5 attribute is set
  to allow the python list to be restored.

Strings
  Python unicode strings are identified using HDF5 attributes.

Note:
  This module was written before the standardization of order preserving
  dictionaries in python 3.6. It is possible that key order saving could be
  now be removed. I am not sure how h5py handles things internally.

"""

import logging
import pathlib
import zipfile
import re

import numpy as np
import h5py
from collections import OrderedDict

log = logging.getLogger(__name__)

def dictToHdf5(dick, filename=None, driver=None, **kwargs):
    """
    Save a python dictionary as an HDF5 data file.

    This will only work for dictionaries where the keys are all strings. The
    order of the dictionary will always be preserved.

    Programming Notes
    -----------------

    In order to retain the order of the input dictionary I create an dictionary
    order attribute as part of the HDF5 root '/'. The h5py library will by default 
    always return the groups before the datasets, so I need to save the original
    order somehow.
    """
    
    # I need to check if the file exists here.  
    # This can also be done by using the correct flags in h5py.File.
    file = h5py.File(filename, 'w', driver=driver)

    if not isinstance(dick, dict):
        raise Exception('Incorrect input type. Dictionary expected.')

    _addDictToHdf5(file, dick, **kwargs)

    file.close()


def hdf5ToDict(filename, driver=None, include=None, exclude=None):
    """
    Read an HDF5 file into a dictionary.

    This will only work for simple HDF files in which we only care
    about the data name and value.

    Keywords
    --------

    exclude
      String or list of strings. Strings must be regular expressions.
      Any keys matching an exclude pattern will be excluded in the
      returned dictionary. 

    include
      String or list of strings. Strings must be regular expressions.
      Any excluded keys matching an include pattern will be included
      in the returned dictionary. If no exclude pattern is given then
      exclude will be set to '.*'. The order of evaluation is exclude
      patterns first, then include patterns.
    """

    include, exclude = _setupIncludeExclude(include, exclude)
    
    file = h5py.File(filename, 'r', driver=driver)

    dick = _createNewItemFromHdf5(file, include=include, exclude=exclude)
    
    file.close()

    return dick


def dictToHdf5Zip(dick, filename=None, **kwargs):
    """
    Write a dictionary into a zipped hdf5 file.
    
    This works by:
      1. writing a hdf5 file.
      2. moving that file into a zip archive.
      3. delete the hdf5 file.

    This will overwrite and delete existing files without any warning!
    """
    
    p = pathlib.Path(filename)
    if p.suffix == '.zip':
        file_z = filename
        file_h = str(p.parent/p.stem)
    else:
        file_z = filename+'.zip'
        file_h = filename

    # is_file_h = pathlib.Path(path_h).exists()
    # is_file_z = pathlib.Path(path_z).exists()
    
    dictToHdf5(dick, filename=file_h, **kwargs)

    with zipfile.ZipFile(
            file_z
            ,mode='w'
            ,compression=zipfile.ZIP_DEFLATED
            ,compresslevel=6
            ) as ff_z:
        ff_z.write(file_h, pathlib.Path(file_h).name)

    pathlib.Path(file_h).unlink()

    
def hdf5ZipToDict(filename, **kwargs):
    """
    This works by:
      1. Extract hdf5 file from zip archive.
      2. Read hdf5 file.
      3. delete the hdf5 file.

    This will overwrite and delete existing files without any warning!
    """

    p = pathlib.Path(filename)
    if p.suffix == '.zip':
        file_z = filename
        file_h = str(p.parent/p.stem)
    else:
        file_z = filename+'.zip'
        file_h = filename

    # is_file_h = pathlib.Path(path_h).exists()
    # is_file_z = pathlib.Path(path_h).exists()

    # Extract the hdf5 file.
    # Here we assume that there is only one file in the zip archive.
    with zipfile.ZipFile(file_z, mode='r') as ff_z:
        # Strip out any path information.
        info = ff_z.infolist()[0]
        info.filename = pathlib.Path(info.filename).name
        ff_z.extract(ff_z.infolist()[0], p.parent)
        
    dick = hdf5ToDict(file_h, **kwargs)
    pathlib.Path(file_h).unlink()

    return dick

    
def _addItemToHdf5(group, key, item, compression=None, compression_opts=None):
    if compression is True: compression = 'lzf'

    try:
        if isinstance(item, dict):
            new_group = group.create_group(key)
            _addDictToHdf5(
                new_group
                ,item
                ,compression=compression
                ,compression_opts=compression_opts)
        elif isinstance(item, list):
            new_group = group.create_group(key)
            _addListToHdf5(
                new_group
                ,item
                ,compression=compression
                ,compression_opts=compression_opts)
        elif item is None:
            group[key] = False
            group[key].attrs['_mirhdf5 python None'] = True
        else:
            if np.isscalar(item):
                group.create_dataset(
                    key
                    ,data=item)
                if isinstance(item, str):
                    group[key].attrs['_mirhdf5 python str'] = True
            else:
                group.create_dataset(
                    key
                    ,data=item
                    ,compression=compression
                    ,compression_opts=compression_opts)
    except TypeError:
        log.exception('Could not add key "{}" of type {} to hdf5 file.'.format(key, type(item)))


def _addDictToHdf5(group, dick, **kwargs):

    keys = [key.encode() for key in dick.keys()]
    group.attrs['_mirhdf5 python object type'] = 'dict'.encode()
    try:
        group.attrs['_mirhdf5 dictionary order'] = keys
    except TypeError:
        log.error('Could not save dictionary key order. keys of unsupported data type.')


    for key in dick:
        _addItemToHdf5(group, key, dick[key], **kwargs)


def _addListToHdf5(group, in_list, **kwargs):

    group.attrs['_mirhdf5 python object type'] = 'list'.encode()

    for ii, item in enumerate(in_list):
        key = '{:04d}'.format(ii)
        _addItemToHdf5(group, key, item, **kwargs)


def _addHdf5ToDict(
        group
        ,dick
        ,attrs=None
        ,group_type=None
        ,include=None
        ,exclude=None
        ):

    if group_type != 'dict':
        raise Exception('HDF5 parsing error. Expected a dict type, got {} instead.'.format(group.attrs['_mirhdf5 python object type']))

    if '_mirhdf5 dictionary order' in attrs:
        key_list = attrs['_mirhdf5 dictionary order']
    else:
        key_list = group.keys()

    for key in key_list:
        # Convert bytes key to approprate Python string type.
        key_string = key
        if hasattr(key, 'decode'):
            key_string = key.decode()
                
        if _keyIncluded(key_string, include, exclude):
            new_item = _createNewItemFromHdf5(group, key, include, exclude)
            dick[key_string] = new_item


def _addHdf5ToList(
        group
        ,in_list
        ,attrs=None
        ,group_type=None
        ,include=None
        ,exclude=None
        ):

    if group_type != 'list':
        raise Exception('HDF5 parsing error. Expected a list type.')
    
    keys = group.keys()
    for key in keys:
        new_item = _createNewItemFromHdf5(group, key, include, exclude)
        in_list.append(new_item)


def _createNewItemFromHdf5(group, key=None, include=None, exclude=None):

    if key is None:
        data = group
    else:
        data = group[key]
    attrs = data.attrs

    if isinstance(data, h5py.Group):
        if '_mirhdf5 python object type' in attrs:
            group_type = attrs['_mirhdf5 python object type']
            if hasattr(group_type, 'decode'):
                group_type = group_type.decode()
        else:
            group_type = 'dict'

        if group_type == 'dict':
            new_item = dict()
            _addHdf5ToDict(data, new_item, attrs, group_type, include, exclude)
        elif group_type == 'list':
            new_item = list()
            _addHdf5ToList(data, new_item, attrs, group_type, include, exclude)
        else:
            raise Exception('Unknown group type: {}'.format(group_type))
    elif '_mirhdf5 python None' in attrs:
        new_item = None
    else:
        new_item = data[()]

        if '_mirhdf5 python str' in attrs:
            new_item = new_item.decode()

    return new_item


def _setupIncludeExclude(include, exclude):
        
    # If a single string was given, turn it into a list.
    # Then compile all the regular expressions.
    if include is not None:
        if isinstance(include, str):
            include = [include]
        include = [re.compile(xx) for xx in include]

        # if only include was provide, then exclude everything else.
        if exclude is None:
            exclude = ['.*']
            
    if exclude is not None:
        if isinstance(exclude, str):
            exclude = [exclude]
        exclude = [re.compile(xx) for xx in exclude]
        
    return include, exclude


def _keyIncluded(key, include, exclude):

    status = True
    if exclude:
        for regex in exclude:
            if regex.match(key):
                status = False
    if include:
        for regex in include:
            if regex.match(key):
                status = True
                
    return status



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>

A set of utility routines to load bragg reflection data files from
several external applications including x0h, XOP and SHADOW.

Data will be returned in a standardized dictionary.
"""

import numpy as np
import os

import logging
log = logging.getLogger(__name__)

def read(filename, filetype=None):
    """
    The main routine to read Bragg reflection data files.

    This will switch between the various format specific routines
    based on the filetype parameter.

    Parameters
    ----------
    filename : str
      The file to read.

    Keyword Arguments
    -----------------
    filetype : str
      The source of the given file. If not provided, the filetype will be
      guessed. Currently supported types are: 'xop', 'x0h'.
    """

    # Try to guess the filetype if it was not provided.
    if filetype is None:
        filetype = _guess_filetype(filename)
        m_log.info(f'Rocking filetype guess: {filetype}')
    if filetype is None:
        raise Exception('Could not guess the filetype. Please use the filetype keyword.')

    if filetype == 'xop':
        out = read_xop(filename)
    elif filetype == 'x0h':
        raise NotImplementedError(f'A reader for filetype {filetype} not yet implemented.')
    else:
        raise Exception(f'Filetype {filetype} not recognized.')

    return out

def _guess_filetype(filename):
    pathname, basename = os.path.split(filename)
    rootname, extname = os.path.splitext(basename)
    if rootname == 'diff_pat':
        filetype = 'xop'
    else:
        filetype = None
    return filetype

def _read_xop__diff_pat_dat__header(filename):
    header = ''
    with open(filename) as ff:
        while True:
            line = ff.readline()
            if line[0] != '#':
                break
            header += line
    return header

def _read_xop__diff_pat_dat__data(filename):
    data = np.loadtxt(filename, dtype=np.float64)
    return data

def read_xop(filename):
    """
    Read a data file from XOP and return a standard rocking-curve dict.

    It is expected that the given file is diff_pat.dat file from XOP.
    """
    pathname, basename = os.path.split(filename)
    rootname, extname = os.path.splitext(basename)
    if not extname == '.dat':
        m_log.warning(f'File extion .dat expected for XOP data. Found {extname} instead.')

    m_log.info(f'Reading XOP data from{filename}')

    header = _read_xop__diff_pat_dat__header(filename)
    data = _read_xop__diff_pat_dat__data(filename)

    col_list = [
        ('dtheta_in',   'urad', 'Th-Th Bragg (in)')
        ,('dtheta_out', 'urad', 'Th-Th Bragg (out)')
        ,('phase_p',    'rad',  'Phase pi')
        ,('phase_s',    'rad',  'Phase sigma')
        ,('circular',   '',     'Circ. Polariz.')
        ,('reflect_p',  '',     'Reflectivity pi')
        ,('reflect_s',  '',     'Reflectivity sigma')
    ]

    out = {}
    out['value'] = {}
    out['units'] = {}
    out['label'] = {}

    for ii, info in enumerate(col_list):
        name, units, label = info
        out['value'][name] = data[:,ii]
        out['units'][name] = units
        out['label'][name] = label

    return out


# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir pablant <npablant@pppl.gov>
"""

from multiprocessing import Pool
from xicsrt.xicsrt_raytrace import *

m_log = mirlogging.getLogger(__name__)

def raytrace(config, processes=None):
    """
    Perform a series of ray tracing runs using the `multiprocessing` cpu pool.

    Each run will rebuild all objects, reset the random seed and then
    perform the requested number of iterations.

    If the option 'save_images' is set, then images will be saved
    at the completion of each run.

    Also see :func:`~xicsrt.xicsrt_raytrace.raytrace` for a single process
    version of this routine.
    """
    profiler.start('raytrace_multiprocessing')
    
    # Update the default config with the user config.
    config = xicsrt_config.get_config(config)
    check_config(config)
    
    num_runs = config['general']['number_of_runs']
    random_seed = config['general']['random_seed']

    mp_result_list = []
    output_list = []

    with Pool(processes) as pool:
        # loop through each configuration in the configuration input file
        # and add a new run into the pool.
        for ii in range(num_runs):
            m_log.info('Adding run to pool: {} of {}'.format(ii + 1, num_runs))

            # Make a copy of the configuration.              
            config_run = deepcopy(config)         
            config_run['general']['output_run_suffix'] = '{:04d}'.format(ii)
            
            # Make sure each run uses a unique random seed.
            if random_seed is not None:
                random_seed += ii
            config_run['general']['random_seed'] = random_seed
                
            arg = (config_run, True)
            mp_result = pool.apply_async(raytrace_single, arg)
            mp_result_list.append(mp_result)
        pool.close()
        pool.join()

    profiler.start('mp: gathering')
    # Gather all the results together.
    for mp_result in mp_result_list:
        output = mp_result.get()
        output_list.append(output)
    profiler.stop('mp: gathering')

    output = combine_raytrace(output_list)

    # Reset the configuration options that were unique to the individual runs.
    output['config']['general']['output_run_suffix'] = config['general']['output_run_suffix']
    output['config']['general']['random_seed'] = config['general']['output_run_suffix']

    if config['general']['save_config']:
        xicsrt_io.save_config(output['config'])
    if config['general']['save_images']:
        xicsrt_io.save_images(output)
    if config['general']['save_results']:
        xicsrt_io.save_results(output)
    if config['general']['print_results']:
        print_raytrace(output)
        
    profiler.stop('raytrace_multiprocessing')
    return output



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>

A set of algorithms to generate vector distributions.

.. Note::
    The term 'spread' here denotes the angular range of emission. The term
    'divergence' is not used because I normally think of a divergence as a
    gaussian distributed probability distribution of angles. This type of
    distribution is available, but for generality and consistency 'spread'
    will be used throughout.
"""

import numpy as np


def vector_distribution(spread, number, name=None):
    """
    A convenience function to retrieve vector distributions by name.

    Parameters
    ----------
    spread : float or array [radians]
      Can be a scalar or an array. See individual distributions for format
      and definitions.

    number : int
      The number of vectors to generate.

    name : string ('isotropic')
      The name of the vector distribution. Available names:
      'isotropic', 'isotropic_xy', 'flat', 'flat_xy', 'gaussian'.

    Returns
    -------
    ndarray
      A numpy array of shape (number, 3) containing the generated unit vectors.
    """
    if name is None: name = 'isotropic'

    name = name.lower()
    if name == 'isotropic':
        func = vector_dist_isotropic
    elif name == 'isotropic_xy':
        func = vector_dist_isotropic_xy
    elif name == 'flat':
        func = vector_dist_flat
    elif name == 'flat_xy':
        func = vector_dist_flat_xy
    elif name == 'gaussian':
        func = vector_dist_gaussian
    else:
        raise Exception(f'Distribution "{name}" is not known.')

    return func(spread, number)

def solid_angle(spread, name=None):
    """
    A convenience function to retrieve solid angles that correspond to
    the various vector distributions.

    Units: [sr]
    """
    if name is None: name = 'isotropic'

    name = name.lower()
    if name == 'isotropic':
        func = solid_angle_isotropic
    elif name == 'isotropic_xy':
        func = solid_angle_isotropic_xy
    else:
        raise Exception(f'Solid angle calculation for "{name}" is not available.')

    return func(spread)

def vector_dist_isotropic(spread, number):
    """
    Return unit vectors from an isotropic (uniform spherical) distribution that
    fall within an angular spread (divergence) of theta.

    The ray cone is aligned along the z-axis.

    Parameters
    ----------
    spread : float [radians]
      The half-angle of the emitted cone of vectors (axis to edge).

    number : int
      The number of vectors to generate.

    Returns
    -------
    ndarray
        A numpy array of shape (number, 3) containing the generated unit vectors.
    """
    theta = _parse_spread_single(spread)

    z = np.random.uniform(np.cos(theta), 1, number)
    phi = np.random.uniform(0, 2*np.pi, number)

    output = np.empty((number, 3))
    output[:, 0] = np.sqrt(1 - z**2) * np.cos(phi)
    output[:, 1] = np.sqrt(1 - z**2) * np.sin(phi)
    output[:, 2] = z

    return output

def solid_angle_isotropic(spread):
    """
    Calculate the solid angle for the vector_dist_isotropic distribution.

    Parameters
    ----------
    spread : float [radians]
      The half-angle of cone of vectors (axis to edge).

    Returns
    -------
    solid_angle
      Units: [sr]
    """
    theta = _parse_spread_single(spread)
    solid_angle = 4 * np.pi * np.sin(theta[0]/2)**2
    return solid_angle

def vector_dist_isotropic_xy(spread, number):
    """
    Return random unit vectors from an isotroptic (uniform spherical) distribution
    that fall within a given x and y angular spread.

    The truncated-cone of vectors is aligned along the z-axis.

    .. Note::
        This routine repeatedly filters from a circular distribution, which is
        accurate but not efficent. Efficiency goes down for more unequal values
        of the x and y spread.

    .. Todo::
        Replace vector_dist_isotropic_xy with a more efficent calculation.
        A possible approach is to calculate the 2D Joint Cumulative
        Distribution Function for isotropic emission on a flat plane.

    Parameters
    ----------
    spread : float or array [radians]
      | The half-angles in the x and y directions that define the extent of the
      | truncated-cone of vectors. Spread can be contain either 1,2 or 4 values.
      |
      | s or [s]
      |   A single value that will be used for both the x and y directions.
      | [x, y]
      |   Two values values that will be used for the x and y directions.
      | [xmin, xmax, ymin, ymax]
      |   For values that define the asymmetric exent in x and y directions.
      |   Example: [-0.1, 0.1, -0.5, 0.5]

    number : int
      The number of vectors to generate.

    Returns
    -------
    ndarray
      A numpy array of shape (number, 3) containing the generated unit vectors.
    """
    theta = _parse_spread_xy(spread)

    # Determine the extent of the circular cone than we need.
    theta_xmax = np.max(np.abs(theta[0:2]))
    theta_ymax = np.max(np.abs(theta[2:]))
    theta_max = np.arcsin(np.sqrt(np.sin(theta_xmax) ** 2 + np.sin(theta_ymax) ** 2))

    # Generate and filter rays until we have the requested number.
    output = np.empty((number, 3))
    n_filled = 0
    while n_filled < number:
        vectors = vector_dist_isotropic(theta_max, number)
        mask = np.ones(number, dtype=bool)

        mask &= vectors[:, 0] / np.sqrt(vectors[:, 0] ** 2 + vectors[:, 2] ** 2) > np.sin(theta[0])
        mask &= vectors[:, 0] / np.sqrt(vectors[:, 0] ** 2 + vectors[:, 2] ** 2) <= np.sin(theta[1])
        mask &= vectors[:, 1] / np.sqrt(vectors[:, 1] ** 2 + vectors[:, 2] ** 2) > np.sin(theta[2])
        mask &= vectors[:, 1] / np.sqrt(vectors[:, 1] ** 2 + vectors[:, 2] ** 2) <= np.sin(theta[3])

        n_new = np.sum(mask)
        n_need = number - n_filled
        if n_new > (number - n_filled):
            n_new = n_need

        output[n_filled:n_filled + n_new, :] = vectors[mask, :][:n_new, :]
        n_filled += n_new

    return output

def solid_angle_isotropic_xy(spread):
    """
    Calculate the solid angle for the vector_dist_isotropic_xy distribution.

    Units: [sr]
    """
    theta = _parse_spread_xy(spread)
    solid_angle = (
        np.arcsin(np.abs(np.sin(theta[0])*np.sin(theta[2])))
        + np.arcsin(np.abs(np.sin(theta[0])*np.sin(theta[3])))
        + np.arcsin(np.abs(np.sin(theta[1])*np.sin(theta[2])))
        + np.arcsin(np.abs(np.sin(theta[1])*np.sin(theta[3])))
        )
    return solid_angle

def vector_dist_flat(spread, number):
    """
    Return unit vectors from an flat (uniform planar) distribution that
    fall within an angular spread.

    The ray cone is aligned along the z-axis.

    Parameters
    ----------
    spread : float [radians]
      The half-angle of the emitted cone of vectors (axis to edge).

    number : int
      The number of vectors to generate.

    Returns
    -------
    ndarray
      A numpy array of shape (number, 3) containing the generated unit vectors.
    """
    theta = _parse_spread_single(spread)

    r = np.sqrt(np.random.uniform(0, np.tan(theta), number))
    angle1 = np.random.uniform(0, 2*np.pi, number)

    angle0 = np.arctan(r)

    output = np.empty((number, 3))
    output[:, 0] = np.cos(angle1) * np.sin(angle0)
    output[:, 1] = np.sin(angle1) * np.sin(angle0)
    output[:, 2] = np.cos(angle0)

    return output

def vector_dist_flat_xy(spread, number):
    """
    Return random unit vectors from an flat (uniform planar) distribution
    that fall within a given x and y angular spread.

    The truncated-cone of vectors is aligned along the z-axis.

    .. Note::
        This distribution is identical to that used by the SHADOW raytracing
        code for both the 'flat' and 'uniform' distributions (as of 2021-01).

    Parameters
    ----------
    spread : float or array [radians]
      | The half-angles in the x and y directions that define the extent of the
      | truncated-cone of vectors. Spread can be contain either 1,2 or 4 values.
      |
      | s or [s]
      |   A single value that will be used for both the x and y directions.
      | [x, y]
      |   Two values values that will be used for the x and y directions.
      | [xmin, xmax, ymin, ymax]
      |   For values that define the asymmetric exent in x and y directions.
      |   Example: [-0.1, 0.1, -0.5, 0.5]

    number : int
      The number of vectors to generate.

    Returns
    -------
    ndarray
      A numpy array of shape (number, 3) containing the generated unit vectors.
    """
    theta = _parse_spread_xy(spread)
    range = np.tan(theta)

    x = np.random.uniform(range[0], range[1], number)
    y = np.random.uniform(range[2], range[3], number)

    angle0 = np.arctan(np.sqrt(x ** 2 + y ** 2))
    angle1 = np.arctan2(y, x)

    output = np.empty((number, 3))
    output[:, 0] = np.cos(angle1) * np.sin(angle0)
    output[:, 1] = np.sin(angle1) * np.sin(angle0)
    output[:, 2] = np.cos(angle0)

    return output


def vector_dist_flat_gaussian(spread, num_samples):
    """
    Create distribution of vectors with a Gaussian distribution on a flat
    plane. The ray code is aligned with the z-axis, so the distribution is
    Gaussian in the x-y directions.

    For small angles this distribution will approximate the Kent distribution
    and will be approximately gaussian in angle.

    Parameters
    ----------
    spread : float [radians]
      The half-with-at-half-max (hwhm) of the Gaussian angular distribution.

    Returns
    -------
    ndarray
      A numpy array of shape (number, 3) containing the generated unit vectors.
    """

    theta = _parse_spread_single(spread)

    # Convert the angluar hwhm to sigma.
    sigma = theta[0] / (np.sqrt(2 * np.log(2)))

    # Calculate the equivilant width in x & y on a plane with z=1.
    xsigma = np.sin(sigma)
    ysigma = np.sin(sigma)

    # Origins for a gaussian distribution of rays.
    mean = [0, 0]

    # Define a diagnonal covariance matrix.
    cov = [[xsigma**2, 0], [0, ysigma**2]]

    x, y = np.random.multivariate_normal(mean, cov, num_samples).T
    z = np.full(num_samples, 1.0)

    out = np.vstack((x, y, z)).T
    out_mod = np.linalg.norm(out, axis=1)
    out = np.einsum('ij,i->ij', out, 1 / np.linalg.norm(out, axis=1), optimize=True)

    return out


def _to_ndarray(spread):
    """
    Convert input value to an numpy array.
    Scalars will be transformed to a one element array.
    """
    if np.isscalar(spread):
        out = np.array([spread])
    elif isinstance(spread, np.ndarray):
        out = spread
    else:
        out = np.array(spread)
    return out

def _parse_spread_single(spread):
    """
    Parse the number of input values in spread and return a standard array.
    Use only for distribution with a single spread value.
    """
    spread = _to_ndarray(spread)
    if len(spread) != 1:
        raise Exception('Spread must be a scalar or one element array.')

    return spread

def _parse_spread_xy(spread):
    """
    Parse the number of input values in spread and return a standard array.
    Use only for assymetric xy distributions.
    """
    spread = _to_ndarray(spread)
    if len(spread) == 1:
        out = [-spread[0], spread[0], -spread[0], spread[0]]
    elif len(spread) == 2:
        out = [-spread[0], spread[0], -spread[1], spread[1]]
    elif len(spread) == 4:
        out = [spread[0], spread[1], spread[2], spread[3]]
    else:
        raise Exception('Spread must have 1, 2 or 3 elements. See docstring.')

    return out



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir A. Pablant <nablant@pppl.gov>
    Yevgeniy Yakusevich <eugenethree@gmail.com>

Description
-----------

A plasma object with toroidal geometry and a circular cross-section.
Here the plasma profiles are read from a text files.
"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.sources._XicsrtPlasmaToroidal import XicsrtPlasmaToroidal
from xicsrt.util import profiler

@dochelper
class XicsrtPlasmaToroidalDatafile(XicsrtPlasmaToroidal):

    def default_config(self):
        config = super().default_config()
        config['emissivity_file']  = None
        config['temperature_file'] = None
        config['velocity_file']    = None
        return config
        
    def get_emissivity(self, rho):
        output = np.zeros(len(rho))
        # Read and interpolate profile from data file
        data  = np.loadtxt(self.param['emissivity_file'], dtype = np.float64)           
        output[:]  = np.interp(rho, data[:,0], data[:,1], left=0.0, right=0.0)
        
        return output

    def get_temperature(self, rho):
        output = np.zeros(len(rho))
        
        # Read and interpolate profile from data file
        data  = np.loadtxt(self.param['temperature_file'], dtype = np.float64)           
        output[:]  = np.interp(rho, data[:,0], data[:,1], left=0.0, right=0.0)
        
        return output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
"""
Authors:
  | Novimir Antoniuk Pablant <npablant@pppl.gov>

Purpose:
  Create a simple profiler module.

Description:
  This module is meant to enable manual profiling with very low overhead.

"""
# ------------------------------------------------------------------------------


import os
import datetime

import logging

log = logging.getLogger(__name__)
profiler_results = dict()
flags = {}
flags['enabled'] = False

def isEnabled():
    return flags['enabled']

def startProfiler(reset=False):
    if reset:
        resetProfiler()
        
    flags['enabled'] = True

def stopProfiler():
    flags['enabled'] = False

    # Reset the start time of every entry.
    # It doesn't make sense to keep timing active when not using the profiler.
    for key in profiler_results:
        profiler_results[key]['time_start'] = None

def resetProfiler():
    global profiler_results
    profiler_results = dict()

def report(flush=True):
    names = list(profiler_results.keys())
    totals = [profiler_results[xx]['time_total'] for xx in profiler_results]
    indexes = sorted(range(len(totals)), key=totals.__getitem__, reverse=True)

    if flush:
        print('', flush=True, end='')

    log.info('{:25.25s} {:14.14s}  {:14.14s}  {:6.6s}'.format(
        'name'
        ,'total'
        ,'single'
        ,'calls'))
    
    for ii in indexes:
        name = names[ii]
        if profiler_results[name]['num_calls'] > 0:
            log.info('{:25.25s} {}  {}  {:6d}'.format(
                name
                ,profiler_results[name]['time_total']
                ,profiler_results[name]['time_total']/profiler_results[name]['num_calls']
                ,profiler_results[name]['num_calls']))

def getTimeTotal(name):
    return profiler_results[name]['time_total']

def getTimeSingle(name):
    return profiler_results[name]['time_total']/profiler_results[name]['num_calls']

def start(name):
    if flags['enabled']:
        if not name in profiler_results:
            _newProfile(name)
            
        profiler_results[name]['time_start'] = datetime.datetime.now()

def stop(name):
    if flags['enabled']:
        if profiler_results[name]['time_start'] is not None:
            profiler_results[name]['time_total'] += datetime.datetime.now() - profiler_results[name]['time_start']
            profiler_results[name]['num_calls'] += 1
            profiler_results[name]['time_start'] = None

def _newProfile(name):
    profiler_results[name] = {
        'time_total':datetime.timedelta(0)
        ,'time_start':None
        ,'num_calls':0
        }

    



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>
    Yevgeniy Yakusevich <eugenethree@gmail.com>

Define the :class:`InteractMosaicCrystal` class.
"""

import numpy as np

from xicsrt.util import profiler

from xicsrt.tools import xicsrt_spread
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractCrystal import InteractCrystal

@dochelper
class InteractMosaicCrystal(InteractCrystal):
    """
    A class to handle Mosaic Crystal optics.

    .. Todo::
      InteractMosaicCrystal efficiency could be improved by including a
      pre-filter. The pre-filter would use a step-function rocking curve to
      exclude rays that are outside the likely range of reflection with the
      current mosaic spread.
    """

    def default_config(self):
        """
        mosaic_spread : float (0.0) [radians]
          The fwhm of the Gaussian distribution of crystalite normals around
          the nominal surface normal.

        mosaic_depth : int (15)
          The number of crystalite layers to model. This value will depend
          on the crystal structure and the incident x-ray energy.

        mosaic_cutoff: float (None)
          A numerical probability cutoff used to avoid calculation of the
          mosaic reflection for angles far away from the nominal Bragg angle.
          A value of 1e-8 would provide a 6-sigma cutoff in angles, while a
          value of 1e-14 would provide an 8-sigma cutoff.
          By default no cutoff is used.
        """
        config = super().default_config()
        config['mosaic_spread'] = 0.0
        config['mosaic_depth'] = 15
        config['mosaic_cutoff'] = None
        return config

    def interact(self, rays, xloc, norm, mask=None):
        """
        Model reflections from a mosaic crystal using a multi-layer model.
        This is meant to simulate the penetration of x-rays into the HOPG until
        the rays either encounter a crystalite that satisfies the Bragg
        condition or get absorbed. This method of calculation replicates both
        the HOPG 'focusing' qualities as well as the expected throughput.
        """
        if mask is None: mask = rays['mask']
        m = mask

        # If a cutoff is given then check for rays whose angle away from the
        # nominal bragg angle falls outside of the cutoff and mark them as
        # not-reflected.
        if self.param['mosaic_cutoff'] is not None:
            bragg_angle, incident_angle = self.angle_calc(rays, norm, mask)
            angle_fwhm = self.param['mosaic_spread']
            angle_sigma = angle_fwhm/(2 * np.sqrt(2 * np.log(2)))
            angle_cutoff = np.sqrt(-1*np.log(self.param['mosaic_cutoff'])*2*angle_sigma**2)
            m[m] = np.abs(bragg_angle[m] - incident_angle[m]) < angle_cutoff

            num_outside = np.sum(np.abs(bragg_angle[m] - incident_angle[m]) > angle_cutoff)
            self.log.debug(f'mosaic angle_cutoff: {np.degrees(angle_cutoff):0.2f} deg, removing {num_outside} rays.')

        # Check if any rays were within the cutoff.
        if np.sum(m) > 0:
            mosaic_mask = np.zeros(rays['mask'].shape, dtype=np.bool_)

            # Check for mosaic reflections up to the given 'depth'.
            profiler.start('mosaic: loop')
            for ii in range(self.param['mosaic_depth']):
                temp_mask = (~ mosaic_mask) & mask
                num_temp = np.sum(temp_mask)
                if num_temp == 0:
                    break

                self.log.debug('  Mosaic iteration: {} rays: {}'.format(ii, num_temp))

                profiler.start('mosaic: mosaic_normals')
                norm_mosaic = self.mosaic_normals(norm, temp_mask)
                profiler.stop('mosaic: mosaic_normals')

                profiler.start('mosaic: angle_check')
                temp_mask = self.angle_check(rays, norm_mosaic, temp_mask)
                profiler.stop('mosaic: angle_check')

                profiler.start('mosaic: reflect_vectors')
                rays = self.reflect_vectors(rays, xloc, norm_mosaic, temp_mask)
                profiler.stop('mosaic: reflect_vectors')

                mosaic_mask[temp_mask] = True
            profiler.stop('mosaic: loop')
            mask &= mosaic_mask

        return rays

    def mosaic_normals(self, normals, mask, copy=True):
        """
        Add mosaic spread to the normals.
        Generates a list of crystallite normal vectors in a Gaussian
        distribution around the nominal surface normals.
        """
        m = mask

        rad_hwhm = self.param['mosaic_spread']/2.0

        profiler.start('mosaic: vector_dist_flat_gaussian')
        dir_local = xicsrt_spread.vector_dist_flat_gaussian(rad_hwhm, np.sum(m))
        profiler.stop('mosaic: vector_dist_flat_gaussian')

        R = np.empty((np.sum(m), 3, 3,), dtype=np.float64)

        # Create two vectors perpendicular to the surface normal and each other,
        # it doesn't matter how they are oriented otherwise.
        R[:, 0, :]  = np.cross(normals[m], [1,0,0]) + np.cross(normals[m], [0,0,1])
        R[:, 0, :] /= np.linalg.norm(R[:, 0, :], axis=1)[:, np.newaxis]
        R[:, 1, :]  = np.cross(normals[m], R[:, 0, :])
        R[:, 1, :] /= np.linalg.norm(R[:, 1, :], axis=1)[:, np.newaxis]
        R[:, 2, :] = normals[m]

        if copy:
            norm_out = normals.copy()
        else:
            norm_out = normals

        norm_out[m] = np.einsum('ij,ijk->ik', dir_local, R)
        return norm_out





# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
"""

# Import top level user functions.
from xicsrt.xicsrt_raytrace import raytrace
from xicsrt.xicsrt_public import *
from xicsrt._version import __version__
from xicsrt.util.version import warn_version

try:
    from xicsrt.xicsrt_multiprocessing import raytrace as raytrace_mp
except:
    raise Warning('multiprocessing could not be loaded on your platform.')



# -*- coding: utf-8 -*-
"""
.. Authors:
   Novimir Pablant <npablant@pppl.gov>

Define the :class:`ShapeMeshcylinder` class.
"""
import numpy as np
from scipy.spatial import Delaunay
from xicsrt.util import profiler
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.tools import xicsrt_math as xm

from xicsrt.optics._ShapeMesh import ShapeMesh

@dochelper
class ShapeMeshCylinder(ShapeMesh):
    """
    A toroidal crystal implemented using a mesh-grid.

    This class meant to be used for three reasons:
    - A toroidal optic shape usable with large radii of curvature
    - As an example and template for how to implement a mesh-grid optic.
    - As a verification of the mesh-grid implementation.

    The analytical :class:`Shapecylinder` object will be much faster.

    **Programming Notes**

    This optic is built in local coordinates with the mesh surface normal
    generally in the local z = [0, 0, 1] direction and with
    config['trace_local'] = True. This is recommended because the mesh
    implementation performs triangulation and interpolation in the local
    x-y plane.

    """

    def default_config(self):
        """
        radius: float (1.0)
            The radius of curvature of the crystal.

        mesh_size : (float, float) ((11,11))
          The number of mesh points in the x and y directions.

        mesh_coarse_size : (float, float) ((5,5))
          The number of mesh points in the x and y directions.
        """
        config = super().default_config()
        config['mesh_refine'] = True
        config['mesh_size'] = (11,11)
        config['mesh_coarse_size'] = (5,5)
        config['mesh_xsize'] = None
        config['mesh_ysize'] = None

        config['radius'] = 1.0

        # It's good practice to always define meshes in local coordinates.
        config['trace_local'] = True

        return config

    def setup(self):
        super().setup()
        self.log.debug('Yo mama was here.')

        # Calculate the angles that define the physical mesh size
        if (xsize := self.param['mesh_xsize']) is None:
            xsize = self.param['xsize']
        if (ysize := self.param['mesh_ysize']) is None:
            ysize = self.param['ysize']
        self.param['x_range'] = [-1*xsize/2, xsize/2]
        half_angle = np.arcsin(ysize/2/(self.param['radius']))
        self.param['angle_range'] = [-1*half_angle, half_angle]

        # Generate the fine mesh.
        mesh_points, mesh_normals, mesh_faces = self.generate_mesh(self.param['mesh_size'])
        self.param['mesh_points'] = mesh_points
        self.param['mesh_normals'] = mesh_normals
        self.param['mesh_faces'] = mesh_faces

        # Generate the coarse mesh.
        mesh_points, mesh_normals, mesh_faces = self.generate_mesh(self.param['mesh_coarse_size'])
        self.param['mesh_coarse_points'] = mesh_points
        self.param['mesh_coarse_normals'] = mesh_normals
        self.param['mesh_coarse_faces'] = mesh_faces

        # Calculate the final width and height of the optic.
        mesh_local = self.param['mesh_points']
        mesh_xsize = np.max(mesh_local[:,0])-np.min(mesh_local[:,0])
        mesh_ysize = np.max(mesh_local[:,1])-np.min(mesh_local[:,1])
        self.log.debug(f"Mesh xsize x ysize: {mesh_xsize:0.3f}x{mesh_ysize:0.3f}")

    def cylinder(self, x, angle):
        """
        Return a 3D surface coordinate given a distance and angle.
        """
        C0 = np.array([0.0, 0.0, 0.0])
        C0_zaxis  = np.array([0.0, 0.0, 1.0])
        C0_xaxis  = np.array([1.0, 0.0, 0.0])
        radius     = self.param['radius']

        x_vec = np.array([x, 0.0, 0.0])

        O = C0 + radius * C0_zaxis + x_vec

        X_norm = xm.vector_rotate(C0_zaxis, C0_xaxis, angle)
        X = O - radius * X_norm

        return X, X_norm

    def shape(self, a, b):
        return self.cylinder(a, b)

    def calculate_mesh(self, a, b):
        profiler.start('calculate_mesh')

        num_a = len(a)
        num_b = len(b)

        aa, bb = np.meshgrid(a, b, indexing='ij')

        xx = np.empty((num_a, num_b))
        yy = np.empty((num_a, num_b))
        zz = np.empty((num_a, num_b))

        normal_xx = np.empty((num_a, num_b))
        normal_yy = np.empty((num_a, num_b))
        normal_zz = np.empty((num_a, num_b))

        # ------------------------------------------------
        # Now calculate the xyz values at each grid point.
        for ii_a in range(num_a):
            for ii_b in range(num_b):

                a = aa[ii_a, ii_b]
                b = bb[ii_a, ii_b]

                xyz, norm = self.shape(a, b)

                xx[ii_a, ii_b] = xyz[0]
                yy[ii_a, ii_b] = xyz[1]
                zz[ii_a, ii_b] = xyz[2]
                normal_xx[ii_a, ii_b] = norm[0]
                normal_yy[ii_a, ii_b] = norm[1]
                normal_zz[ii_a, ii_b] = norm[2]

        profiler.stop('calculate_mesh')

        return xx, yy, zz, normal_xx, normal_yy, normal_zz

    def generate_mesh(self, mesh_size=None):
        """
        This method creates the meshgrid for the crystal
        """
        profiler.start('generate_mesh')

        # --------------------------------
        # Setup the basic grid parameters.

        a_range = self.param['x_range']
        b_range = self.param['angle_range']

        num_a = mesh_size[0]
        num_b = mesh_size[1]

        self.log.debug(f'num_a, num_b: {num_a}, {num_b}, total: {num_a*num_b}')

        a = np.linspace(a_range[0], a_range[1], num_a)
        b = np.linspace(b_range[0], b_range[1], num_b)

        xx, yy, zz, normal_xx, normal_yy, normal_zz = \
            self.calculate_mesh(a, b)

        aa, bb = np.meshgrid(a, b, indexing='ij')
        angles_2d = np.stack((aa.flatten(), bb.flatten()), axis=0).T

        points = np.stack((xx.flatten(), yy.flatten(), zz.flatten())).T
        normals = np.stack((normal_xx.flatten(), normal_yy.flatten(), normal_zz.flatten())).T

        delaunay = Delaunay(angles_2d)
        faces = delaunay.simplices

        # It's also possible to triangulate using the x,y coordinates.
        # This is not recommended unless there is some specific need.
        #
        # delaunay = Delaunay(points[:, 0:2])
        # faces = delaunay.simplices


        profiler.stop('generate_mesh')
        return points, normals, faces




"""
A module for dealing with colors.

This module is an extension/modification of parts of:
  matplotlib.colors
  matplotlib.cm
  matplotlib.pyplot

There are a number of things that I don't like about how the matplotlib
versions handle things, so this is my attempt to correct some of these
issues.

Maybe someday I'll try to tackle the matplotlib code directly and create
a replacement for the original code.

Example
-------

  norm = mircolor.Normalize(0.0, 1.0)
  grad = mircolor.getColorGradient(norm, 'tab10')
  color = grad.to_rgba(1.0)

"""

import numpy as np
from matplotlib import cm
from matplotlib import colors
from matplotlib.colors import Normalize


defined_gradients = {
    'mirred':{
        'type':'LinearSegmentedColorGradient'
        ,'red':[(0.0, 0.5, 0.5)
                ,(0.5, 1.0, 1.0)
                ,(1.0, 1.0, 1.0)]
        ,'green':[(0.0, 0.0, 0.0)
                  ,(0.5, 0.0, 0.0)
                  ,(1.0, 0.8, 0.8)]
        ,'blue':[(0.0, 0.0, 0.0)
                 ,(0.5, 0.0, 0.0)
                 ,(1.0, 0.8, 0.8)]}
    ,'mirgreen':{
        'type':'LinearSegmentedColorGradient'
        ,'green':[(0.0,1.0,1.0)
                  ,(1.0,0.0,0.0)]}
    ,'mirblue':{
        'type':'LinearSegmentedColorGradient'
        ,'red':[(0.0, 0.0, 0.0)
                ,(0.2, 0.0, 0.0)
                ,(0.4, 0.44, 0.44)
                ,(0.6, 0.44, 0.44)
                ,(0.8, 0, 0.0)
                ,(1.0, 0, 0.0)]
        ,'green':[(0.0, 0.16, 0.16)
                  ,(0.2, 0, 0.0)
                  ,(0.4, 0.6, 0.6)
                  ,(0.6, 0.75, 0.75)
                  ,(0.8, 0.62, 0.62)
                  ,(1.0, 0.41, 0.41)]
        ,'blue':[(0.0, 0.63, 0.63)
                 ,(0.2, 1.0, 1.0)
                 ,(0.4, 0.8, 0.8)
                 ,(0.6, 0.8, 0.8)
                 ,(0.8, 1.0, 1.0)
                 ,(1.0, 0.54, 0.54)]}
    ,'mirRedBlackBlue':{
        'type':'LinearSegmentedColorGradient'
        ,'red':[(0.0, 0.8, 0.8)
                ,(0.5, 0.0, 0.0)
                ,(1.0, 0.0, 0.0)]
        ,'green':[(0.0, 0.0, 0.0)
                  ,(0.5, 0.0, 0.0)
                  ,(1.0, 0.8, 0.8)]
        ,'blue':[(0.0, 0.0, 0.0)
                 ,(0.5, 0.0, 0.0)
                 ,(1.0, 0.8, 0.8)]}
    # mirrainbow is not finished.
    ,'mirrainbow':{
        'type':'LinearSegmentedColorGradient'
        ,'red':[(0.0, 1.0, 1.0)
                ,(0.25, 1.0, 1.0)
                ,(0.5, 0.0, 0.0)
                ,(0.75, 0.0, 0.0)
                ,(1.0, 0.0, 0.0)]
        ,'green':[(0.0, 0.0, 0.0)
                ,(0.25, 0.0, 0.0)
                ,(0.5, 1.0, 1.0)
                ,(0.75, 1.0, 1.0)
                ,(1.0, 0.0, 0.0)]
        ,'blue':[(0.0, 0.0, 0.0)
                ,(0.25, 0.0, 0.0)
                ,(0.5, 0.0, 0.0)
                ,(0.7, 0.0, 0.0)
                ,(1.0, 1.0, 1.0)]}    
}


def getColorGradient(norm=None, cmap=None):
    if not cmap in defined_gradients:

        if not cmap in cm.__dict__.keys():
            raise NameError('Gradient {} not defined'.format(cmap))

        return cm.ScalarMappable(norm, cmap)

    if defined_gradients[cmap]['type'] == 'LinearSegmentedColorGradient':
        gradient = LinearSegmentedColorGradient(norm, defined_gradients[cmap])
    else:
        raise Exception('Gradient {} has wrong type.  Not currently supported'.format(cmap))

    return gradient
    
    
class ColorGradient(object):

    def __call__(self, *args):
        return self.to_rgba(*args)

    
class LinearSegmentedColorGradient(ColorGradient):

    rgba_keys = ['red', 'blue', 'green', 'alpha']
            
    def __init__(self, norm=None, segmentdata=None):
        if norm:
            self.norm = norm
        else:
            self.norm = Normalize()
            
        self.setSegmentData(segmentdata)

    
    def _getRgbaComp(self, value, key):

        data = self._segmentdata[key]
        x = data[:, 0]
        y0 = data[:, 1]
        y1 = data[:, 2]

        if value <= x[0]:
            output = y0[0]
        elif value >= x[-1]:
            output = y0[-1]
        else:
            ii = np.searchsorted(x, value)-1

            output = (value - x[ii])/(x[ii+1] - x[ii]) * (y1[ii+1] - y0[ii]) + y0[ii]
            
        return output

    
    def _scalarToRgba(self, value):
        
        return (self._getRgbaComp(self.norm(value), 'red')
                ,self._getRgbaComp(self.norm(value), 'green')
                ,self._getRgbaComp(self.norm(value), 'blue')
                ,self._getRgbaComp(self.norm(value), 'alpha'))

    
    def _arrayToRgba(self, values):
        
        output = np.zeros(values.shape+(4,))

        it = np.nditer(values, flags=['multi_index'])
        while not it.finished:
            output[it.multi_index] = self._scalarToRgba(values[it.multi_index])
            it.iternext()

        return output

    
    def to_rgba(self, value, alpha=None):
               
        if np.isscalar(value):
            output = self._scalarToRgba(value)
        else:
            output = self._arrayToRgba(value)

        if alpha is not None:
            output[:, 3] = alpha

        return output


    def setSegmentData(self, segmentdata_in):

        self._segmentdata = {}

        for kk in self.rgba_keys:
            if kk in segmentdata_in:
                self._segmentdata[kk] = self._standardizeSegmentData(segmentdata_in[kk])
            elif kk == 'alpha':
                # Set default values for the alpha channel.
                default_data = [[0.0, 1.0, 1.0],[1.0, 1.0, 1.0]]
                self._segmentdata[kk] = self._standardizeSegmentData(default_data)
            else:
                # Set default values for the red, green and blue channels.
                default_data = [[0.0, 0.0, 0.0],[1.0, 0.0, 0.0]]
                self._segmentdata[kk] = self._standardizeSegmentData(default_data)

                
    def _standardizeSegmentData(self, data):

        try:
            array_data = np.array(data)
        except:
            raise TypeError("segment data must be convertable to an array")        

        if not (len(array_data.shape) == 2 and array_data.shape[1] == 3):
            raise ValueError("segment data must be a nx3 array.")

        x = array_data[:, 0]
        if x[0] != 0. or x[-1] != 1.0:
            raise ValueError(
                "segment data points must start with x=0. and end with x=1")
        if np.sometrue(np.sort(x) - x):
            raise ValueError(
                "segment data points must have x in increasing order")

        return array_data



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>

A set of tools to facilitate string handling in XICSRT.
"""

import numpy as np

def simplify_strings(value):
    """
    Recursively simplify strings in the given variable.

    This will do the following:

      1. Make all strings lower case.
    """
    if isinstance(value, str):
        value = str.lower(value)
    elif isinstance(value, dict):
        for key in value:
            value[key] = simplify_strings(value[key])
    elif isinstance(value, list):
        for ii in range(len(value)):
            value[ii] = simplify_strings(value[ii])
    elif isinstance(value, np.ndarray):
        for ii in range(len(value)):
            value[ii] = simplify_strings(value[ii])
    return value



# -*- coding: utf-8 -*-
"""
.. Authors:
   Novimir Pablant <npablant@pppl.gov>

To prepare an update:
  1. Update _version.py
  2. git tag -a v0.4.0 -m "Version 0.4.0"
  3. git push --tags
  4. Make everything is merged with 'master' and 'stable'.
  5. Push to both bitbucket and github.

To submit an update to pypi:
  1. python setup.py sdist bdist_wheel
  2. python -m twine upload --repository testpypi dist/*
  3. python -m twine upload dist/*
"""

import setuptools

with open('README.md') as ff:
    long_description = ff.read()

exec(open('xicsrt/_version.py').read())

classifiers = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3",
    "Operating System :: OS Independent",
    ]

params ={
    'name': 'xicsrt',
    'version': __version__,
    'author': 'Novimir Antoniuk Pablant',
    'author_email': 'npablant@pppl.gov',
    'maintainer': 'Novimir Antoniuk Pablant',
    'maintainer_email': 'npablant@pppl.gov',
    'description': 'A photon based raytracing application written in Python.',
    'long_description': long_description,
    'long_description_content_type': 'text/markdown',
    'url': 'https://github.com/PrincetonUniversity/xicsrt',
    'license': 'MIT',
    'packages': setuptools.find_packages(exclude=['xicsrt_contrib*']),
    'entry_points': {'console_scripts':['xicsrt=xicsrt.__main__:run']},
    'classifiers': classifiers,
    'install_requires': ['numpy', 'scipy', 'pillow', 'h5py'],
    'python_requires': '>=3.8',
    }
    
setuptools.setup(**params)



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`XicsrtOpticMeshCylindricalCrystal` class.
"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractCrystal import InteractCrystal
from xicsrt.optics._ShapeMeshCylinder import ShapeMeshCylinder

@dochelper
class XicsrtOpticMeshCylindricalCrystal(InteractCrystal, ShapeMeshCylinder):
    """
    A Cylindrical crystal optic implemented using a mesh.
    """

    pass



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>

These are a set of routines for 3D visualization using the
ipyvolume library.

.. warning::

    This API for this module is currently out of date. The plotly module
    currently has the latest API implementation.

Example
-------
Example code for using this module within a Jupyter notebook.

.. code-block::

    import xicsrt.visual.xicsrt_3d__ipyvolume as xicsrt_3d

    xicsrt_3d.figure()
    xicsrt_3d.add_rays(results)
    xicsrt_3d.add_optics(config)
    xicsrt_3d.add_sources(config)
    xicsrt_3d.show()
"""

import ipyvolume as ipv
import numpy as np

import matplotlib

from xicsrt import xicsrt_config
from xicsrt.objects._Dispatcher import Dispatcher

def truncate_mask(mask, max_num):
    num_mask = np.sum(mask)
    if num_mask > max_num:
        w = np.where(mask)[0]
        np.random.shuffle(w)
        mask[w[max_num:]] = False

    return mask

def figure():
    fig = ipv.figure(width=900, height=900)
    return fig

def add_rays(results):
    config = results['config']
    _add_ray_history(results['found']['history'], config, lost=False)
    _add_ray_history(results['lost']['history'], config, lost=True)

def _add_ray_history(history, config, lost=None):

    if lost is False:
        color = [1.0, 0.0, 0.0, 0.1]
    elif lost is True:
        color = [0.0, 0.0, 1.0, 0.1]
    else:
        color = [0.0, 0.0, 0.0, 0.1]
        
    num_elem = len(history)
    key_list = list(history.keys())
    
    for ii in range(num_elem-1):

        # All rays leaving this optic element.
        mask = history[key_list[ii]]['mask']

        # This is a temporary solution to filter lost rays for
        # which no intersection at the next optic was calculated.
        #
        # This is not a good solution overall since it is possible
        # to imagine a case where we would want to retain rays at
        # the origin.
        mask &= history[key_list[ii+1]]['origin'][:, 0] != 0.0

        num_mask = np.sum(mask)

        if num_mask == 0:
            continue
        
        x0 = history[key_list[ii]]['origin'][mask, 0]
        y0 = history[key_list[ii]]['origin'][mask, 1]
        z0 = history[key_list[ii]]['origin'][mask, 2]
        x1 = history[key_list[ii+1]]['origin'][mask, 0]
        y1 = history[key_list[ii+1]]['origin'][mask, 1]
        z1 = history[key_list[ii+1]]['origin'][mask, 2]

        x = np.concatenate((x0, x1))
        y = np.concatenate((y0, y1))
        z = np.concatenate((z0, z1))

        lines = np.zeros((num_mask, 2), dtype=int)
        lines[:, 0] = np.arange(0, num_mask)
        lines[:, 1] = np.arange(num_mask, num_mask * 2)
        
        obj = ipv.plot_trisurf(x, y, z, lines=lines, color=color)
        obj.line_material.transparent = True
        obj.line_material.linewidth = 10.0
    

def generate_colors(history, config):
    """
    This is some old code to generate colors for the ray plotting
    in _add_ray_history.

    For the moment I am just saving the old code here, but this
    will not work with the new history dictionary. (Only minor 
    adjustments are neccessary to make it work, but the 
    implementation was not very general.)
    """
    raise NotImplementedError()


    color_masks = []
    color_list = []

    # This is a good pink color for un-reflected rays.
    # color_unref = [1.0, 0.5, 0.5, 0.2]

    if False:
        color_masks.append(output[0]['mask'].copy())
        color_list.append((1.0, 0.0, 0.0, 0.5))

    # Color by vertical position in Plasma.
    if False:
        norm = matplotlib.colors.Normalize(0.0, 1.0)
        cm = matplotlib.cm.ScalarMappable(norm=norm, cmap='gist_rainbow')

        color_masks.append((output[0]['origin'][:, 2] > 0.1))
        color_list.append(cm.to_rgba(0.0, alpha=0.5))

        color_masks.append((output[0]['origin'][:, 2] > 0.05) & (output[0]['origin'][:, 2] < 0.1))
        color_list.append(cm.to_rgba(1 / 5, alpha=0.5))

        color_masks.append((output[0]['origin'][:, 2] > 0.0) & (output[0]['origin'][:, 2] < 0.05))
        color_list.append(cm.to_rgba(2 / 5, alpha=0.5))

        color_masks.append((output[0]['origin'][:, 2] > -0.05) & (output[0]['origin'][:, 2] < 0))
        color_list.append(cm.to_rgba(3 / 5, alpha=0.5))

        color_masks.append((output[0]['origin'][:, 2] > -0.1) & (output[0]['origin'][:, 2] < -0.05))
        color_list.append(cm.to_rgba(4 / 5, alpha=0.5))

        color_masks.append((output[0]['origin'][:, 2] < -0.1))
        color_list.append(cm.to_rgba(5 / 5, alpha=0.5))

    if True:
        norm = matplotlib.colors.Normalize(0.0, 1.0)
        cm = matplotlib.cm.ScalarMappable(norm=norm, cmap='gist_rainbow')

        delta = 0.005
        num_wave = 5
        for ii in range(num_wave):
            wave_0 = inputs['source_input']['wavelength'] - 1 * delta * (num_wave - 1) / 2 + delta * ii
            wave_min = wave_0 - delta / 2
            wave_max = wave_0 + delta / 2
            mask_temp = (output[0]['wavelength'][:] > wave_min) & (output[0]['wavelength'][:] < wave_max)
            print('{:7.4f} {:7.4f} {:7.4f} {}'.format(wave_0, wave_min, wave_max, np.sum(mask_temp)))
            color_masks.append(mask_temp)
            color_list.append(cm.to_rgba(ii / num_wave, alpha=0.5))

def add_surf(obj):
    w = obj.param['xsize'] / 2.0
    h = obj.param['ysize'] / 2.0
    d = obj.param['zsize'] / 2.0

    points = np.zeros((8, 3))
    points[0, :] = [w, h, d]
    points[1, :] = [w, -h, d]
    points[2, :] = [-w, h, d]
    points[3, :] = [-w, -h, d]

    points[4, :] = [w, h, -d]
    points[5, :] = [w, -h, -d]
    points[6, :] = [-w, h, -d]
    points[7, :] = [-w, -h, -d]

    points_ext = obj.point_to_external(points)

    x = points_ext[:, 0]
    y = points_ext[:, 1]
    z = points_ext[:, 2]

    # I am sure there is a way to automate this using a meshgrid,
    # but at the moment this is faster.
    if d > 0:
        triangles = (
            (0, 3, 1)
            , (0, 3, 2)

            , (4, 7, 5)
            , (4, 7, 6)

            , (0, 6, 2)
            , (0, 6, 4)

            , (0, 5, 1)
            , (0, 5, 4)

            , (3, 6, 2)
            , (3, 6, 7)

            , (3, 5, 1)
            , (3, 5, 7)
        )
    else:
        triangles = (
            (0, 3, 1)
            , (0, 3, 2)
        )

    ipv_obj = ipv.plot_trisurf(x, y, z, triangles=triangles, color=[1.0, 1.0, 0.0, 0.2])
    ipv_obj.material.transparent = True
    
def add_optics(config):

    for key_opt in config['optics']:
        config_opt = config['optics'][key_opt]
        config_opt = xicsrt_config.config_to_numpy(config_opt)

        if True:
            w = config_opt['width'] / 2.0
            h = config_opt['height'] / 2.0
            cx = config_opt['xaxis']
            cy = np.cross(config_opt['xaxis'], config_opt['zaxis'])

            point0 = w * cx + h * cy + config_opt['origin']
            point1 = h * cy + config_opt['origin']
            point2 = -1 * w * cx + h * cy + config_opt['origin']

            point3 = w * cx + config_opt['origin']
            point4 = config_opt['origin']
            point5 = -1 * w * cx + config_opt['origin']

            point6 = w * cx - h * cy + config_opt['origin']
            point7 = -1 * h * cy + config_opt['origin']
            point8 = -1 * w * cx - h * cy + config_opt['origin']

            points = np.array([
                point0
                , point1
                , point2
                , point3
                , point4
                , point5
                , point6
                , point7
                , point8
            ])

            x = points[:, 0]
            y = points[:, 1]
            z = points[:, 2]

            # I am sure there is a way to automate this using a meshgrid,
            # but at the moment this is faster.
            triangles = (
                (4, 0, 2)
                , (4, 2, 8)
                , (4, 6, 8)
                , (4, 6, 0)
            )

            obj = ipv.plot_trisurf(x, y, z, triangles=triangles, color=[0.5, 0.5, 0.5, 0.5])
            obj.material.transparent = True

        if 'crystal' in str.lower(key_opt) and 'radius' in config_opt:

            # Initialize the configuration.
            # This merges the user config that we created
            # with the default config
            config = xicsrt_config.get_config(config)

            # The easiest way to do coordinate transformations is
            # to use the raytrace objects, since they already have
            # everything built in.  Of course we could do this
            # by just using some simple matrix multiplications,
            # but this way we can use existing code.
            #
            # We could just instantiate the objects directly as needed,
            # but since a dispatcher is already available in xicsrt that
            # does this for us, we might as well use it.  This is copied
            # from xicsrt_raytrace.raytrace_single.
            name = 'crystal'
            section = 'optics'
            optics = Dispatcher(config, section)
            optics.instantiate()
            optics.setup()
            optics.initialize()

            # Get the crystal object from the dispatcher.
            optic_obj = optics.get_object(name)

            crystal_center_ext = config_opt['origin'] + config_opt['zaxis'] * config_opt['radius']
            crystal_center_loc = optic_obj.point_to_local(crystal_center_ext)

            x = np.array([crystal_center_ext[0]])
            y = np.array([crystal_center_ext[1]])
            z = np.array([crystal_center_ext[2]])
            ipv.scatter(x, y, z, color='black', marker="sphere")

            # Plot the crystal circle.
            num = 1000
            crystal_radius = config_opt['radius']
            coord_loc = np.zeros((num, 3))
            coord_loc[:, 0] = np.sin(np.linspace(0.0, np.pi * 2, num)) * config_opt['radius'] + crystal_center_loc[0]
            coord_loc[:, 1] = crystal_center_loc[1]
            coord_loc[:, 2] = np.cos(np.linspace(0.0, np.pi * 2, num)) * config_opt['radius'] + crystal_center_loc[2]
            coord_ext = optic_obj.point_to_external(coord_loc)
            x = coord_ext[:,0]
            y = coord_ext[:,1]
            z = coord_ext[:,2]
            lines = np.zeros((num, 2), dtype=int)
            lines[:, 0] = np.arange(num)
            lines[:, 1] = np.roll(lines[:, 0], 1)
            obj = ipv.plot_trisurf(x, y, z, lines=lines, color=[0.0, 0.0, 0.0, 0.5])

            rowland_center_ext = config_opt['origin'] + config_opt['zaxis'] * config_opt['radius'] / 2
            rowland_center_loc = optic_obj.point_to_local(rowland_center_ext)
            rowland_radius = crystal_radius / 2
            coord_loc = np.zeros((num, 3))
            coord_loc[:, 0] = np.sin(np.linspace(0.0, np.pi * 2, num)) * rowland_radius + rowland_center_loc[0]
            coord_loc[:, 1] = rowland_center_loc[1]
            coord_loc[:, 2] = np.cos(np.linspace(0.0, np.pi * 2, num)) * rowland_radius + rowland_center_loc[2]
            coord_ext = optic_obj.point_to_external(coord_loc)
            x = coord_ext[:,0]
            y = coord_ext[:,1]
            z = coord_ext[:,2]
            lines = np.zeros((num, 2), dtype=int)
            lines[:, 0] = np.arange(num)
            lines[:, 1] = np.roll(lines[:, 0], 1)
            obj = ipv.plot_trisurf(x, y, z, lines=lines, color=[0.0, 0.0, 0.0, 0.5])
            
def add_sources(config):
    # Update the default config with the user config.
    config = xicsrt_config.get_config(config)
    
    # Combine the user and default object pathlists.
    pathlist = []
    pathlist.extend(config['general']['pathlist_objects'])
    pathlist.extend(config['general']['pathlist_default'])
    
    sources = Dispatcher(config, 'sources')
    sources.instantiate()
    sources.setup()
    sources.initialize()

    for key in config['sources']:
        obj = sources.get_object(key)
        
    add_surf(obj)

def show():
    view = [0,0,0]
    ipv.xlim(view[1] - 0.5, view[1] + 0.5)
    ipv.ylim(view[1] - 0.5, view[1] + 0.5)
    ipv.zlim(view[2] - 0.5, view[2] + 0.5)

    #ipv.style.axes_off()
    #ipv.style.box_off()

    ipv.show()



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`InteractObject` class.
"""

import numpy as np
from copy import deepcopy

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._TraceObject import TraceObject
from xicsrt.tools import xicsrt_math as xm

@dochelper
class InteractObject(TraceObject):
    """
    The base class for interactions of rays with surfaces in XICSRT.

    This base class should be used to define behavior such as reflection,
    transmission and absorption.
    """

    def interact(self, rays, xloc, norm, mask=None):
        """
        Evaluate interaction with a surface.
        The base-class has no interaction, rays just pass through.
        """
        O = rays['origin']

        if mask is not None:
            rays['mask'][:] = mask

        # The following line does two things:
        # 1. Set the origin for rays (lost or found) with an intersection.
        # 2. Set the rays without an intersection to nan.
        O[:] = xloc[:]

        return rays



"""
Version Description:
major.minor.revision

major:
  - Major changes in codebase or functionality.

minor:
  - Changes to the public API
  - Old config files can no longer be loaded without conversion.
  - New output dictionaries may break existing external code.

revision:
  - New release that does not break any compatibility.
"""

__version__="0.8.12"
__version_date__ = "2024-08-20"



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>

Contains the XicsrtPlasmaGeneric class.
"""
import logging

import numpy as np

from xicsrt.util import profiler
from xicsrt.tools import xicsrt_spread
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.objects._GeometryObject import GeometryObject
from xicsrt.sources._XicsrtSourceFocused import XicsrtSourceFocused

@dochelper
class XicsrtPlasmaGeneric(GeometryObject):
    """
    A generic plasma object.

    Plasma object will generate a set of ray bundles where each ray bundle
    has the properties of the plasma at one particular real-space point.

    Each bundle is modeled by a SourceFocused object.

    .. Note::
      If a `voxel` type bundle is used rays may be generated outside of the
      defined plasma volume (as defined by xsize, ysize and zsize). The bundle
      *centers* are randomly distributed throughout the plasma volume, but this
      means that if a bundle is (randomly) placed near the edges of the plasma
      then the bundle voxel volume may extend past the plasma boundary. This
      behavior is expected. If it is important to have a sharp plasma boundary
      then consider using the 'point' bundle_type instead.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filter_objects = []

    def default_config(self):
        """
        xsize
          The size of this element along the xaxis direction.

        ysize
          The size of this element along the yaxis direction.

        zsize
          The size of this element along the zaxis direction.

        angular_dist : string ('isotropic')
          The type of angular distribution to use for the emitted rays.
          Available distributions: 'isotropic', 'isotropic_xy', 'flat',
          'flat_xy', 'gaussian', and 'gaussian_flat'.
          See `XicsrtSourceGeneric` for documentation of each distribution.

          Warning: Only the 'isotropic' distribution is currently supported!

        spread: float (None) [radians]
          The angular spread for the emission cone. The spread defines the
          half-angle of the cone. See 'angular_dist' in :any:`XicsrtSourceGeneric`
          for detailed documentation.

        spread_radius: float (None) [meters]
          If specified, the spread will be calculated for each bundle such that
          the spotsize at the target matches the given radius. This is useful
          when working with very extended plasma sources.
          This options is incompatible with 'spread'.

        use_poisson
          No documentation yet. Please help improve XICSRT!

        wavelength_dist : string ('voigt')
          No documentation yet. Please help improve XICSRT!

        wavelength : float (1.0) [Angstroms]
          No documentation yet. Please help improve XICSRT!

        mass_number : float (1.0) [au]
          No documentation yet. Please help improve XICSRT!

        linewidth : float (0.0) [1/s]
          No documentation yet. Please help improve XICSRT!

        emissivity : float (0.0) [ph/m^3]
          No documentation yet. Please help improve XICSRT!

        temperature : float (0.0) [eV]
          No documentation yet. Please help improve XICSRT!

        velocity : float (0.0) [m/s]
          No documentation yet. Please help improve XICSRT!

        time_resolution : float (1e-3) [s]
          No documentation yet. Please help improve XICSRT!

        bundle_type : string ('voxel')
          Define how the origin of rays within the bundle should be distributed.
          Available options are: 'voxel' or 'point'.

        bundle_volume : float (1e-3) [m^3]
          The volume in which the rays within the bundle should distributed.
          if bundle_type is 'point' this will not affect the distribution,
          though it will still affect the number of bundles if bundle_count
          is set to None.

        bundle_count : int (None)
          The number of bundles to generate. If set to `None` then this number
          will be automatically determined by volume/bundle_volume. This default
          means that each bundle represents exactly the given `bundle_volume` in
          the plasma. For high quality raytracing studies this value should
          generally be set to a value much larger than volume/bundle_volume!

        max_rays : int (1e7)
          No documentation yet. Please help improve XICSRT!

        max_bundles : int (1e7)
          No documentation yet. Please help improve XICSRT!

        filters
          No documentation yet. Please help improve XICSRT!

        """
        config = super().default_config()
                
        config['xsize']          = 0.0
        config['ysize']         = 0.0
        config['zsize']          = 0.0

        config['angular_dist']      = 'isotropic'
        config['spread']           = None
        config['spread_radius']    = None
        config['target']           = None
        config['use_poisson']      = False

        config['wavelength_dist']  = 'voigt'
        config['wavelength']       = 1.0
        config['wavelength_range'] = None
        config['mass_number']      = 1.0
        config['linewidth']        = 0.0

        config['emissivity']      = 0.0
        config['temperature']     = 0.0
        config['velocity']        = 0.0

        config['time_resolution'] = 1e-3
        config['bundle_type']     = 'voxel'
        config['bundle_volume']   = 1e-6
        config['bundle_count']    = None
        config['max_rays']        = int(1e7)
        config['max_bundles']     = int(1e7)
        
        config['filters']         = []
        return config

    def initialize(self):
        super().initialize()
        if self.param['max_rays'] is not None:
            self.param['max_rays']     = int(self.param['max_rays'])
        self.param['volume']       = self.config['xsize'] * self.config['ysize'] * self.config['zsize']

        if self.param['bundle_count'] is None:
            self.param['bundle_count'] = self.param['volume']/self.param['bundle_volume']
        self.param['bundle_count'] = int(np.round(self.param['bundle_count']))
        if self.param['bundle_count'] < 1:
            raise Exception(f'Bundle volume is larger than the plasma volume.')
        if self.param['bundle_count'] > self.param['max_bundles']:
            raise ValueError(
                f"Current settings will produce too many bundles ({self.param['bundle_count']:0.2e}). "
                f"Increase the bundle_volume, explicitly set bundle_count or increase max_bundles.")

    def setup_bundles(self):
        self.log.debug('Starting setup_bundles')
        if self.param['bundle_type'] == 'point':
            self.param['voxel_size'] = 0.0
        elif self.param['bundle_type'] == 'voxel':
            self.param['voxel_size'] = self.param['bundle_volume'] ** (1/3)

        # These values should be overwritten in a derived class.
        bundle_input = {}
        bundle_input['origin']       = np.zeros([self.param['bundle_count'], 3], dtype = np.float64)
        bundle_input['temperature']  = np.ones([self.param['bundle_count']], dtype = np.float64)
        bundle_input['emissivity']   = np.ones([self.param['bundle_count']], dtype = np.float64)
        bundle_input['velocity']     = np.zeros([self.param['bundle_count'], 3], dtype = np.float64)
        bundle_input['mask']         = np.ones([self.param['bundle_count']], dtype = np.bool_)
        bundle_input['spread']       = np.zeros([self.param['bundle_count']], dtype = np.float64)
        bundle_input['solid_angle']  = np.zeros([self.param['bundle_count']], dtype = np.float64)
        
        # randomly spread the bundles around the plasma box
        offset = np.zeros((self.param['bundle_count'], 3))
        offset[:,0] = np.random.uniform(-1 * self.param['xsize'] /2, self.param['xsize'] /2, self.param['bundle_count'])
        offset[:,1] = np.random.uniform(-1 * self.param['ysize']/2, self.param['ysize']/2, self.param['bundle_count'])
        offset[:,2] = np.random.uniform(-1 * self.param['zsize'] /2, self.param['zsize'] /2, self.param['bundle_count'])

        bundle_input['origin'][:] = self.point_to_external(offset)

        # Setup the bundle spread and solid angle.
        bundle_input = self.setup_bundle_spread(bundle_input)

        return bundle_input

    def setup_bundle_spread(self, bundle_input):
        """
        Calculate the spread and solid angle for each bundle.

        If the config option 'spread_radius' is provide the spread will be
        determined for each bundle by a spotsize at the target.

        Note: Even if the idea of a spread radius is added to the generic
              source object we still need to calculate and save the results
              here so that we can correctly calcuate the bundle intensities.
        """
        if self.param['spread_radius'] is not None:
            vector = bundle_input['origin'] - self.param['target']
            dist = np.linalg.norm(vector, axis=1)
            spread = np.arctan(self.param['spread_radius']/dist)
        else:
            spread = self.param['spread']

        bundle_input['spread'][:] = spread

        # For the time being the fuction solid_angle is not vectorized, so a
        # loop is necessary.
        for ii in range(len(bundle_input['spread'])):
            bundle_input['solid_angle'][ii] = xicsrt_spread.solid_angle(bundle_input['spread'][ii])

        return bundle_input

    def get_emissivity(self, rho):
        return self.param['emissivity']

    def get_temperature(self, rho):
        return self.param['temperature']

    def get_velocity(self, rho):
        return self.param['velocity']

    def bundle_generate(self, bundle_input):
        self.log.debug('Starting bundle_generate')
        return bundle_input

    def bundle_filter(self, bundle_input):
        self.log.debug('Starting bundle_filter')
        for filter in self.filter_objects:
            bundle_input = filter.filter(bundle_input)
        return bundle_input
    
    def create_sources(self, bundle_input):
        """
        Generate rays from a list of bundles.

        bundle_input
          a list containing dictionaries containing the locations, emissivities,
          temperatures and velocitities and of all ray bundles to be emitted.
        """

        rays_list = []
        count_rays_in_bundle = []

        m = bundle_input['mask']

        # Check if the number of rays generated will exceed max ray limits.
        # This is only approximate since poisson statistics may be in use.

        predicted_rays = int(np.sum(
            bundle_input['emissivity'][m]
            * self.param['time_resolution']
            * self.param['bundle_volume']
            * bundle_input['solid_angle'][m] / (4 * np.pi)
            * self.param['volume']
            / (self.param['bundle_count'] * self.param['bundle_volume'])))

        self.log.debug(f'Predicted rays: {predicted_rays:0.2e}')

        if self.param['max_rays']:
            if predicted_rays > self.param['max_rays']:
                raise ValueError(
                    f"Current settings will produce too many rays ({predicted_rays:0.2e}). "
                    f"Please reduce integration time or adjust other parameters.")

        # Bundle generation loop
        for ii in range(self.param['bundle_count']):
            if not bundle_input['mask'][ii]:
                continue
            profiler.start("Ray Bundle Generation")
            source_config = dict()
            
            # Specially dependent parameters
            source_config['origin']      = bundle_input['origin'][ii]
            source_config['temperature'] = bundle_input['temperature'][ii]
            source_config['velocity']    = bundle_input['velocity'][ii]
            source_config['spread']      = bundle_input['spread'][ii]

            # Calculate the total number of photons to launch from this bundle
            # volume. Since the source can use poisson statistics, this should
            # be of floating point type.
            intensity = (bundle_input['emissivity'][ii]
                         * self.param['time_resolution']
                         * self.param['bundle_volume']
                         * bundle_input['solid_angle'][ii] / (4 * np.pi))
            
            # Scale the number of photons based on the number of bundles.
            #
            # Ultimately we allow bundle_volume and bundle_count to be
            # independent, which means that a bundle representing a volume in
            # the plasma can be launched from virtual volume of a different
            # size.
            #
            # In order to allow this while maintaining overall photon statistics
            # from the plasma, we normalize the intensity so that each bundle
            # represents a volume of plasma_volume/bundle_count.
            #
            # In doing so bundle_volume cancels out, but I am leaving the
            # calculation separate for clarity.
            intensity *= self.param['volume'] / (self.param['bundle_count'] * self.param['bundle_volume'])
            
            source_config['intensity'] = intensity

            # constants
            source_config['xsize']            = self.param['voxel_size']
            source_config['ysize']            = self.param['voxel_size']
            source_config['zsize']            = self.param['voxel_size']
            source_config['zaxis']            = self.param['zaxis']
            source_config['xaxis']            = self.param['xaxis']
            source_config['target']           = self.param['target']
            source_config['mass_number']      = self.param['mass_number']
            source_config['wavelength_dist']  = self.param['wavelength_dist']
            source_config['wavelength']       = self.param['wavelength']
            source_config['wavelength_range'] = self.param['wavelength_range']
            source_config['linewidth']        = self.param['linewidth']
            source_config['angular_dist']      = self.param['angular_dist']
            source_config['use_poisson']      = self.param['use_poisson']
                
            #create ray bundle sources and generate bundled rays
            source       = XicsrtSourceFocused(source_config)
            bundled_rays = source.generate_rays()

            rays_list.append(bundled_rays)
            count_rays_in_bundle.append(len(bundled_rays['mask']))

            profiler.stop("Ray Bundle Generation")

        profiler.start('Ray Bundle Collection')
        # append bundled rays together to form a single ray dictionary.    
        # create the final ray dictionary
        total_rays = int(np.sum(count_rays_in_bundle))
        rays                = dict()
        rays['origin']      = np.zeros((total_rays,3), dtype=np.float64)
        rays['direction']   = np.zeros((total_rays,3), dtype=np.float64)
        rays['wavelength']  = np.zeros((total_rays), dtype=np.float64)
        rays['weight']      = np.zeros((total_rays), dtype=np.float64)
        rays['mask']        = np.ones((total_rays), dtype=np.bool_)

        index = 0
        for ii, num_rays in enumerate(count_rays_in_bundle):
            rays['origin'][index:index+num_rays] = rays_list[ii]['origin']
            rays['direction'][index:index+num_rays] = rays_list[ii]['direction']
            rays['wavelength'][index:index+num_rays] = rays_list[ii]['wavelength']
            rays['weight'][index:index+num_rays] = rays_list[ii]['weight']
            rays['mask'][index:index+num_rays] = rays_list[ii]['mask']
            index += num_rays
        profiler.stop('Ray Bundle Collection')
            
        if len(rays['mask']) == 0:
            raise ValueError('No rays generated. Check plasma input parameters')

        self.log.debug('Bundles Generated:       {:0.4e}'.format(
            len(m[m])))
        self.log.debug('Rays per bundle, mean:   {:0.0f}'.format(
            np.mean(count_rays_in_bundle)))
        self.log.debug('Rays per bundle, median: {:0.0f}'.format(
            np.median(count_rays_in_bundle)))
        self.log.debug('Rays per bundle, max:    {:0d}'.format(
            np.max(count_rays_in_bundle)))
        self.log.debug('Rays per bundle, min:    {:0d}'.format(
            np.min(count_rays_in_bundle)))

        return rays

    def generate_rays(self):
        ## Create an empty list of ray bundles
        bundle_input = self.setup_bundles()
        ## Apply filters to filter out ray bundles
        bundle_input = self.bundle_filter(bundle_input)  
        ## Populate that list with ray bundle parameters, like emissivity
        bundle_input = self.bundle_generate(bundle_input)
        ## Use the list to generate ray sources
        rays = self.create_sources(bundle_input)
        return rays



# -*- coding: utf-8 -*-
"""
.. Authors:
   Novimir Pablant <npablant@pppl.gov>

Define the :class:`ShapeMeshTorus` class.
"""
import numpy as np
from scipy.spatial import Delaunay
from xicsrt.util import profiler
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.tools import xicsrt_math as xm

from xicsrt.optics._ShapeMesh import ShapeMesh

@dochelper
class ShapeMeshTorus(ShapeMesh):
    """
    A toroidal crystal implemented using a mesh-grid.

    This class meant to be used for three reasons:
    - A toroidal optic shape usable with large radii of curvature
    - As an example and template for how to implement a mesh-grid optic.
    - As a verification of the mesh-grid implementation.

    The analytical :class:`ShapeTorus` object will be much faster.

    **Programming Notes**

    This optic is built in local coordinates with the mesh surface normal
    generally in the local z = [0, 0, 1] direction and with
    config['trace_local'] = True. This is recommended because the mesh
    implementation performs triangulation and interpolation in the local
    x-y plane.

    """

    def default_config(self):
        """
        radius_major: float (1.0)
            The radius of curvature of the crystal in the toroidal (xaxis)
            direction. This is not the same as the geometric major radius of the
            axis of a toroid, which in our case would be r_major-r_minor.

        radius_minor: float (0.1)
            The radius of the curvature of the crystal in the poloidal (yaxis)
            direction. This is the same as the geometric minor radius of a
            toroid.

        normal_method: str ('analytic')
            Specify how to calculate the normal vectors at each of the grid
            points.  Supported values are 'analytic' and 'fd'.

            When set to 'fd' a finite difference method will be used. This is
            primarily here as an example for cases in which the surface position
            is easy to calculate but where surface normals are difficult. A
            better option in these cases however is to use auto-differentiation
            to calculate analytical derivatives (for example using jax).

        mesh_size : (float, float) ((11,11))
          The number of mesh points in the x and y directions.

        mesh_coarse_size : (float, float) ((5,5))
          The number of mesh points in the x and y directions.
        """
        config = super().default_config()
        config['mesh_refine'] = True
        config['mesh_size'] = (11,11)
        config['mesh_coarse_size'] = (5,5)
        config['mesh_xsize'] = None
        config['mesh_ysize'] = None

        config['radius_major'] = 1.0
        config['radius_minor'] = 0.2
        config['convex'] = [False, False]

        config['normal_method'] = 'analytic'

        # The meshgrid is defined in local coordinates.
        config['trace_local'] = True

        return config

    def setup(self):
        super().setup()
        self.log.debug('Yo mama was here.')

        if np.all(self.param['convex'] == [False, False]):
            self.param['torus_sign_major'] = 1
            self.param['torus_sign_minor'] = 1
        elif np.all(self.param['convex'] == [False, True]):
            self.param['torus_sign_major'] = 1
            self.param['torus_sign_minor'] = -1
        elif np.all(self.param['convex'] == [True, False]):
            self.param['torus_sign_major'] = -1
            self.param['torus_sign_minor'] = 1
        elif np.all(self.param['convex'] == [True, True]):
            self.param['torus_sign_major'] = -1
            self.param['torus_sign_minor'] = -1
        else:
            raise Exception(f"Cannot be parse convex config option: {self.param['convex']}")


        # Calculate the angles that define the physical mesh size
        if (xsize := self.param['mesh_xsize']) is None:
            xsize = self.param['xsize']
        if (ysize := self.param['mesh_ysize']) is None:
            ysize = self.param['ysize']
        half_major = np.arcsin(xsize/2/(self.param['radius_major']))
        half_minor = np.arcsin(ysize/2/self.param['radius_minor'])
        self.param['angle_major'] = [-1*half_major, half_major]
        self.param['angle_minor'] = [-1*half_minor, half_minor]

        # Generate the fine mesh.
        mesh_points, mesh_normals, mesh_faces = self.generate_mesh(self.param['mesh_size'])
        self.param['mesh_points'] = mesh_points
        self.param['mesh_normals'] = mesh_normals
        self.param['mesh_faces'] = mesh_faces

        # Generate the coarse mesh.
        mesh_points, mesh_normals, mesh_faces = self.generate_mesh(self.param['mesh_coarse_size'])
        self.param['mesh_coarse_points'] = mesh_points
        self.param['mesh_coarse_normals'] = mesh_normals
        self.param['mesh_coarse_faces'] = mesh_faces

        # Calculate final width and height of the optic for debugging.
        mesh_local = self.param['mesh_points']
        mesh_xsize = np.max(mesh_local[:,0])-np.min(mesh_local[:,0])
        mesh_ysize = np.max(mesh_local[:,1])-np.min(mesh_local[:,1])
        self.log.debug(f"Mesh xsize x ysize: {mesh_xsize:0.3f}x{mesh_ysize:0.3f}")

    def torus(self, a, b):
        """
        Return a 3D surface coordinate given a set of two angles.
        """
        C0_zaxis = np.asarray([0.0, 0.0, 1.0])
        C0_xaxis = np.asarray([1.0, 0.0, 0.0])
        s_maj  = self.param['torus_sign_major']
        s_min  = self.param['torus_sign_minor']
        r_maj = self.param['radius_major']
        r_min = self.param['radius_minor']

        C0_yaxis = np.cross(C0_zaxis, C0_xaxis)

        # r_maj is the radius of curvature of the surface, so the torus
        # center is always defined only by this value.
        center = r_maj*C0_zaxis*s_maj

        # This is the projection of point in the x-z plane.
        C_norm = xm.vector_rotate(C0_zaxis, C0_yaxis, a)
        C = center - r_maj * C_norm * s_maj

        # Find the associated point on the torus axis
        Q = C + r_min*C_norm*s_min

        axis = np.cross(C_norm * s_min, C0_yaxis)
        X_norm = xm.vector_rotate(C_norm * s_min, axis, b)
        X = Q - X_norm * r_min

        return X, X_norm

    def shape(self, a, b):
        return self.torus(a, b)

    def shape_fd(self, a, b, delta=None):
        profiler.start('finite difference')
        if delta is None: delta = 1e-8

        xyz, _ = self.torus(a, b)
        xyz1, _ = self.torus(a + delta, b)
        xyz2, _ = self.torus(a, b + delta)

        vec1 = xyz1 - xyz
        vec2 = xyz2 - xyz
        norm_fd = np.cross(vec1, vec2)
        norm_fd /= np.linalg.norm(norm_fd)
        profiler.stop('finite difference')
        return xyz, norm_fd

    def shape_jax(self, a, b):
        raise NotImplementedError()

    def calculate_mesh(self, a, b):
        profiler.start('calculate_mesh')

        num_a = len(a)
        num_b = len(b)

        aa, bb = np.meshgrid(a, b, indexing='ij')

        xx = np.empty((num_a, num_b))
        yy = np.empty((num_a, num_b))
        zz = np.empty((num_a, num_b))

        normal_xx = np.empty((num_a, num_b))
        normal_yy = np.empty((num_a, num_b))
        normal_zz = np.empty((num_a, num_b))

        # ------------------------------------------------
        # Now calculate the xyz values at each grid point.
        for ii_a in range(num_a):
            for ii_b in range(num_b):

                a = aa[ii_a, ii_b]
                b = bb[ii_a, ii_b]

                # Temporary for development.
                if self.param['normal_method'] == 'analytic':
                    xyz, norm = self.shape(a, b)
                elif self.param['normal_method'] == 'fd':
                    xyz, norm = self.shape_fd(a, b)
                elif self.param['normal_method'] == 'jax':
                    xyz, norm = self.shape_jax(a, b)
                else:
                    raise Exception(f"normal_method {self.param['normal_method']} unknown.")

                xx[ii_a, ii_b] = xyz[0]
                yy[ii_a, ii_b] = xyz[1]
                zz[ii_a, ii_b] = xyz[2]
                normal_xx[ii_a, ii_b] = norm[0]
                normal_yy[ii_a, ii_b] = norm[1]
                normal_zz[ii_a, ii_b] = norm[2]

        profiler.stop('calculate_mesh')

        return xx, yy, zz, normal_xx, normal_yy, normal_zz

    def generate_mesh(self, mesh_size=None):
        """
        This method creates the meshgrid for the crystal
        """
        profiler.start('generate_mesh')

        # --------------------------------
        # Setup the basic grid parameters.

        a_range = self.param['angle_major']
        b_range = self.param['angle_minor']

        num_a = mesh_size[0]
        num_b = mesh_size[1]

        self.log.debug(f'num_a, num_b: {num_a}, {num_b}, total: {num_a*num_b}')

        a = np.linspace(a_range[0], a_range[1], num_a)
        b = np.linspace(b_range[0], b_range[1], num_b)

        xx, yy, zz, normal_xx, normal_yy, normal_zz = \
            self.calculate_mesh(a, b)

        aa, bb = np.meshgrid(a, b, indexing='ij')
        angles_2d = np.stack((aa.flatten(), bb.flatten()), axis=0).T

        points = np.stack((xx.flatten(), yy.flatten(), zz.flatten())).T
        normals = np.stack((normal_xx.flatten(), normal_yy.flatten(), normal_zz.flatten())).T

        delaunay = Delaunay(angles_2d)
        faces = delaunay.simplices

        # It's also possible to triangulate using the x,y coordinates.
        # This does not work well for the toroidal shape.
        #
        # delaunay = Delaunay(points[:, 0:2])
        # faces = delaunay.simplices


        profiler.stop('generate_mesh')
        return points, normals, faces




# -*- coding: utf-8 -*-
"""
Authors
-------
  - Novimir Pablant <npablant@pppl.gov>

"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.sources._XicsrtSourceGeneric import XicsrtSourceGeneric

@dochelper
class XicsrtSourceDirected(XicsrtSourceGeneric):
    """
    An extended rectangular ray source with rays emitted in a preferred
    direction.

    This is similar to the SourceGeneric except that an explicit direction
    can be provided instead of always emitting rays along the z-axis.

    This is different from a SourceFocused in that the emission cone is
    always aimed in a fixed direction for every location in the source.
    The SourceFocused instead aims the emission cone at a specific target
    so that the aiming direction changes for different locations within
    the source.
    """

    def default_config(self):
        """
        direction
          The direction in which to emit rays. This direction will define the
          center of the emission code with angular spread `spread`.
        """
        config = super().default_config()
        config['direction'] = None
        return config

    def initialize(self):
        super().initialize()
        # Setup a default direction along the zaxis.
        if self.param['direction'] is None:
            self.param['direction'] = self.param['zaxis']

    def make_normal(self):
        array = np.empty((self.param['intensity'], 3))
        array[:] = self.param['direction']
        normal = array / np.linalg.norm(array, axis=1)[:, np.newaxis]
        return normal



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
"""
import numpy as np
import logging

import importlib
import glob
import os

from copy import deepcopy
import importlib.util

from xicsrt.util import profiler

class Dispatcher():
    """
    A class to help find, initialize and then dispatch calls to
    raytracing objects.

    A dispatcher is used within XICSRT to find and instantiate objects based
    on their specification within the config dictionary. These objects are
    then tracked within the dispatcher, allowing methods to be called on all
    objects sequentially.
    """

    def __init__(self, config=None, section=None):
        self.log = logging.getLogger(self.__class__.__name__)

        self.config = config
        self.section = section

        pathlist = []
        pathlist.extend(config['general'].get('pathlist', []))
        pathlist.extend(config['general'].get('pathlist_default', []))
        self.pathlist = pathlist
        
        self.objects = dict()
        self.meta = dict()
        self.image = dict()
        self.history = dict()

    def instantiate(self, names=None):
        if names is None:
            names = self.config[self.section].keys()
        elif isinstance(names, str):
            names = [names]

        strict = self.config['general']['strict_config_check']

        obj_info = self.find_xicsrt_objects(self.pathlist)
        # self.log.debug(obj_info)

        for key in names:
            obj = self._instantiate_single(
                obj_info
                ,self.config[self.section][key]
                ,strict=strict)
            self.objects[key] = obj

    def find_xicsrt_objects(self, pathlist):
        """
        Return a dictionary with all the XICSRT objects found in the given
        list of paths. Objects are identified by looking for python files
        that start with '_Xicsrt' prefix.

        Programming Notes
        -----------------
        If a given path does not exist glob will just return and empty list.
        For this reason no path existence checking is needed (unless we want
        to raise a user friendly error).
        """

        filepath_list = []
        name_list = []

        for pp in pathlist:
            filepath_list.extend(glob.glob(os.path.join(pp, '_Xicsrt*.py')))

        for ff in filepath_list:
            filename = os.path.basename(ff)
            objectname = os.path.splitext(filename)[0]
            objectname = objectname[1:]
            name_list.append(objectname)

        output = dict()
        for ii, ff in enumerate(name_list):
            output[ff] = {
                'filepath': filepath_list[ii],
                'name': name_list[ii]
            }

        return output

    def _instantiate_single(self, obj_info, config, strict=None):
        """
        Instantiate an object from a list of filenames and a class name.
        """

        if config['class_name'] in obj_info:
            info = obj_info[config['class_name']]
        else:
            raise Exception('Could not find {} in available objects.'.format(config['class_name']))

        spec = importlib.util.spec_from_file_location(info['name'], info['filepath'])
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        cls = getattr(mod, info['name'])
        obj = cls(config, initialize=False, strict=strict)

        return obj

    def get_object(self, name):
        if not name in self.objects:
            self.instantiate(name)
        return self.objects[name]

    def check_config(self, *args, **kwargs):
        for key, obj in self.objects.items():
            obj.check_config(*args, **kwargs)

    def check_param(self, *args, **kwargs):
        for key, obj in self.objects.items():
            obj.check_param(*args, **kwargs)

    def get_config(self, *args, **kwargs):
        config = dict()
        for key, obj in self.objects.items():
            config[key] = obj.get_config(*args, **kwargs)
        return config

    def setup(self, *args, **kwargs):
        for key, obj in self.objects.items():
            obj.setup(*args, **kwargs)

    def initialize(self, *args, **kwargs):
        for key, obj in self.objects.items():
            obj.initialize(*args, **kwargs)

    def generate_rays(self, keep_meta=None, keep_history=None):
        """
        Generates rays from all sources.
        """
        if keep_meta is None: keep_meta = True
        if keep_history is None: keep_history = False
            
        if len(self.objects) == 0:
            raise Exception('No ray sources defined.')
        elif not len(self.objects) == 1:
            raise NotImplementedError('Multiple ray sources are not currently supported.')
                
        for key, obj in self.objects.items():
            rays = obj.generate_rays()

            if keep_meta:
                self.meta[key] = dict()
                self.meta[key]['num_out'] = np.sum(rays['mask'])
            
            if keep_history:
                self.history[key] = deepcopy(rays)
                
        return rays
            
    def trace(self, rays, keep_meta=None, keep_history=None, keep_images=None):
        """
        Perform raytracing for each object in sequence.
        """
        if keep_meta is None: keep_meta = True
        if keep_history is None: keep_history = False
        if keep_images is None: keep_images = False

        profiler.start('Dispatcher: raytrace')

        for key, obj in self.objects.items():
            
            profiler.start('Dispatcher: trace_global')
            rays = obj.trace_global(rays)
            profiler.stop('Dispatcher: trace_global')

            if keep_meta:
                self.meta[key] = dict()
                self.meta[key]['num_out'] = np.sum(rays['mask'])
            
            if keep_history:
                self.history[key] = deepcopy(rays)

            if keep_images:
                profiler.start('Dispatcher: collect')
                self.image[key] = obj.make_image(rays)
                profiler.stop('Dispatcher: collect')
                
        profiler.stop('Dispatcher: raytrace')
        
        return rays
    
    def apply_filters(self, filters):
        # Used by dispatcher objects to apply filters.
        # 'filters' is a filter dispatcher object that contains filter objects
        # 'self' should be a source or optics dispatcher object
        
        # read the filter list for each source and dispatch the matching filters
        for key in self.objects:
            if not 'filters' in self.objects[key].config:
                break
            if self.objects[key].config['filters'] is None:
                break
            for filter_name in filters.objects:
                if filter_name in self.objects[key].config['filters']:
                    self.objects[key].filter_objects.append(filters.objects[filter_name])




# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>

These are a set of routines for 3D visualization using the plotly library.

Example
-------
Example code for using this module within a Jupyter notebook.

.. code-block::

    import xicsrt.visual.xicsrt_3d__plotly as xicsrt_3d

    xicsrt_3d.figure()
    xicsrt_3d.add_rays(results)
    xicsrt_3d.add_optics(config)
    xicsrt_3d.add_sources(config)
    xicsrt_3d.show()
"""

import numpy as np
import scipy
import plotly.graph_objects as go
import matplotlib

from xicsrt import xicsrt_public
from xicsrt.objects._Dispatcher import Dispatcher

# A module level variable that contains the last defined fig object.
# This is for convenience during interactive scripting.
m_figure = None

def plot(results, **kwargs):
    """
    Create a 3d plot using default options.

    Any keywords provided will be passed to `add_rays`. For more control over
    plotting options it is recommended to perform the plotting steps manually as
    shown by the example in the `xicsrt_3d` module docstring.
    """
    fig = figure()
    add_rays(results, figure=fig, **kwargs)
    add_optics(results['config'], figure=fig)
    add_sources(results['config'], figure=fig)
    add_fluxsurfaces(results['config'], figure=fig)
    show(figure=fig)


def _thin_mask(mask, max_num):
    """
    Reduce the number of True elements in a mask array.

    Ray thinning is done randomly. Used to reduce the number of
    rays plotted.
    """
    num_mask = np.sum(mask)
    if num_mask > max_num:
        w = np.where(mask)[0]
        np.random.shuffle(w)
        mask[w[max_num:]] = False

    return mask


def figure(showbackground=False, visible=False):
    global m_figure

    layout = {
        'scene':{
            'aspectmode':'data',
            'xaxis':{
                'showbackground':showbackground,
                'visible':visible,
                },
            'yaxis':{
                'showbackground':showbackground,
                'visible':visible,
                },
            'zaxis':{
                'showbackground':showbackground,
                'visible':visible,
                },
            },
        'paper_bgcolor':'rgba(0,0,0,0)',
        'plot_bgcolor':'rgba(0,0,0,0)',
        }
    fig = go.Figure(layout=layout)
    m_figure = fig

    return fig


def show(figure=None):
    global m_figure
    if figure is None: figure = m_figure

    figure.show()

def add_rays(
        results,
        lost=True,
        found=True,
        **kwargs):
    if lost:
        _plot_ray_history(results['lost']['history'], _lost=True, **kwargs)
    if found:
        _plot_ray_history(results['found']['history'], _lost=False, **kwargs)

def _plot_ray_history(
        history,
        _lost=None,
        figure=None,
        color=None,
        lost_color=None,
        found_color=None,
        #alpha=None,
        #lost_alpha=None,
        #found_alpha=None,
        ):

    global m_figure
    if figure is None: figure = m_figure

    #print(lost_color, found_color)

    if _lost is False:
        if color is None:
            color = 'rgba(255, 0, 0, 0.01)'
        if found_color is not None:
            color = found_color
        name = 'found'
    elif _lost is True:
        if color is None:
            color = 'rgba(0, 0, 255, 0.01)'
        if lost_color is not None:
            color = lost_color
        name = 'lost'
    else:
        if color is None:
            color = 'rgba(0, 0, 0, 0.1)'
        name = 'other'

    color_plotly = _make_plotly_color(color, alpha=None)

    num_elem = len(history)
    key_list = list(history.keys())
    
    for ii in range(num_elem-1):

        # All rays leaving this optic element.
        mask = history[key_list[ii]]['mask']

        # Filter rays for which there is no intersection at the next optic.
        mask &= np.all(history[key_list[ii+1]]['origin'][:, :] != np.nan, axis=1)

        num_mask = np.sum(mask)

        if num_mask == 0:
            continue
        
        x0 = history[key_list[ii]]['origin'][mask, 0]
        y0 = history[key_list[ii]]['origin'][mask, 1]
        z0 = history[key_list[ii]]['origin'][mask, 2]
        x1 = history[key_list[ii+1]]['origin'][mask, 0]
        y1 = history[key_list[ii+1]]['origin'][mask, 1]
        z1 = history[key_list[ii+1]]['origin'][mask, 2]
        nan = np.zeros(num_mask) * np.nan

        x = np.dstack((x0, x1, nan)).flatten()
        y = np.dstack((y0, y1, nan)).flatten()
        z = np.dstack((z0, z1, nan)).flatten()

        line = {'color':color_plotly}
        data = go.Scatter3d(
            x=x
            ,y=y
            ,z=z
            ,mode='lines'
            ,connectgaps=False
            ,line=line
            ,name=name
            ,showlegend=False)
        figure.add_trace(data)


def _add_trace_volume(obj, figure, name=None, opacity=0.5):
    global m_figure
    if figure is None: figure = m_figure

    w = obj.param['xsize']
    if w is None: w = 0.0
    h = obj.param['ysize']
    if h is None: h = 0.0
    d = obj.param['zsize']
    if d is None: d = 0.0

    h /= 2.0
    w /= 2.0
    d /= 2.0

    points = np.zeros((8, 3))
    points[0, :] = [w, h, d]
    points[1, :] = [w, -h, d]
    points[2, :] = [-w, h, d]
    points[3, :] = [-w, -h, d]

    points[4, :] = [w, h, -d]
    points[5, :] = [w, -h, -d]
    points[6, :] = [-w, h, -d]
    points[7, :] = [-w, -h, -d]

    points_ext = obj.point_to_external(points)

    x = points_ext[:, 0]
    y = points_ext[:, 1]
    z = points_ext[:, 2]

    # Define all of the bounding points.
    if d > 0:
        tri = np.array((
            (0, 3, 1)
            , (0, 3, 2)

            , (4, 7, 5)
            , (4, 7, 6)

            , (0, 6, 2)
            , (0, 6, 4)

            , (0, 5, 1)
            , (0, 5, 4)

            , (3, 6, 2)
            , (3, 6, 7)

            , (3, 5, 1)
            , (3, 5, 7)
        ))
    else:
        tri = np.array((
            (0, 3, 1)
            , (2, 3, 0)
        ))


    trace = go.Mesh3d(
        x=x
        ,y=y
        ,z=z
        ,i=tri[:, 0]
        ,j=tri[:, 1]
        ,k=tri[:, 2]
        ,flatshading=True
        ,opacity=opacity
        ,name=name)

    figure.add_trace(trace)

def _add_trace_mesh(obj, figure=None, name=None):
    """
    Add a meshgrid to the 3D plot.
    """
    global m_figure
    if figure is None: figure = m_figure

    faces = obj.param['mesh_faces']
    points = obj.param['mesh_points']

    if obj.param['trace_local']:
        points = obj.point_to_external(points)

    if True:
        # Plot gridlines
        lines = []
        for f in faces:
            lines.append([f[0], f[1]])
            lines.append([f[0], f[2]])
            lines.append([f[1], f[2]])
        lines = np.array(lines)
        lines = np.unique(lines, axis=0)

        x0 = points[lines[:, 0], 0]
        y0 = points[lines[:, 0], 1]
        z0 = points[lines[:, 0], 2]
        x1 = points[lines[:, 1], 0]
        y1 = points[lines[:, 1], 1]
        z1 = points[lines[:, 1], 2]
        nan = np.zeros(len(x0)) * np.nan

        # Add nans between each line which Scatter3D will use to define linebreaks.
        x = np.dstack((x0, x1, nan)).flatten()
        y = np.dstack((y0, y1, nan)).flatten()
        z = np.dstack((z0, z1, nan)).flatten()

        trace = go.Scatter3d(
            x=x
            , y=y
            , z=z
            , mode='lines'
            , line={'color': 'black'}
            , connectgaps=False
            , showlegend=False)
        figure.add_trace(trace)

    if True:
        # Plot surface
        norm = matplotlib.colors.Normalize(np.min(points[:, 2]), np.max(points[:, 2]))
        cm = matplotlib.cm.ScalarMappable(norm=norm, cmap='plasma')
        color = cm.to_rgba(points[:, 2], alpha=0.75)

        trace = go.Mesh3d(
            x=points[:, 0]
            , y=points[:, 1]
            , z=points[:, 2]
            , i=faces[:, 0]
            , j=faces[:, 1]
            , k=faces[:, 2]
            , vertexcolor=color
            , flatshading=True
            , opacity=0.75)

        figure.add_trace(trace)


def add_optics(config, figure=None, **kwargs):
    section = 'optics'
    for name in config[section]:
        add_object(config, name, section, figure=figure, **kwargs)


def add_sources(config, figure=None, **kwargs):
    section = 'sources'
    for name in config[section]:
        add_object(config, name, section, figure=figure, **kwargs)


def add_fluxsurfaces(config, figure=None, **kwargs):
    """
    Plot the 3D flux surfaces of a plasma source.
    This should work for any source object that has a 'car_from_flx' method.
    """
    section = 'sources'
    for name in config[section]:
        _add_fluxsurf_single(config, name, section, figure=figure, **kwargs)


def add_object(config, name, section, figure=None, **kwargs):

    # Use the dispatcher to instantiate and initialize objects.
    optics = Dispatcher(config, section)
    optics.instantiate(name)
    optics.setup()
    optics.initialize()

    # Get the crystal object from the dispatcher.
    obj = optics.get_object(name)

    plot_mesh = 'mesh_points' in obj.param

    if plot_mesh:
        _add_trace_mesh(obj, figure, name, **kwargs)
    else:
        _add_trace_volume(obj, figure, name, **kwargs)


def _gen_fluxsurface_mesh(obj, s, range_m=None, range_n=None):
    """
    Generate points on a flux surface.
    The given input object must have a method 'car_from_flx'.
    """
    if range_m is None: range_m = (0, 2*np.pi)
    if range_n is None: range_n = (0, np.pi/4)

    num_m = 101
    num_n = 101

    num_points = num_m*num_n
    flx = np.empty((num_points, 3))

    val_m = np.linspace(range_m[0], range_m[1], num_m, endpoint=True)
    val_n = np.linspace(range_n[0], range_n[1], num_n, endpoint=True)

    num_points = num_m*num_n
    flx = np.empty((num_points, 3))

    val_mm, val_nn = np.meshgrid(val_m, val_n, indexing='ij')
    flx[:, 0] = s
    flx[:, 1] = val_mm.flatten()
    flx[:, 2] = val_nn.flatten()

    # This should be callable without a loop, but for now leave this as is
    # to support the VMEC stelltools module.
    car = np.empty(flx.shape)
    for ii in range(flx.shape[0]):
        car[ii, :] = obj.car_from_flx(flx[ii, :])

    return flx, car


def _add_fluxsurf_single(
        config,
        name,
        section=None,
        figure=None,
        alpha=None,
        flatshading=None,
        **kwargs):
    """
    Plot the 3D flux surfaces of a plasma source.
    This should work for any object that has a 'car_from_flx' method.
    """
    global m_figure
    if figure is None: figure = m_figure

    if alpha is None: alpha = 0.5
    if flatshading is None: flatshading = True

    obj = xicsrt_public.get_element(config, name, section)

    # Check to see if this object defines fluxspace through a car_from_flux method.
    if not hasattr(obj, 'car_from_flx'):
        return

    norm = matplotlib.colors.Normalize(0.0, 1.0)
    cm = matplotlib.cm.ScalarMappable(norm=norm, cmap='plasma_r')

    rho_array = np.linspace(0.0, 1.0, 10, endpoint=True)
    for ii, rho in enumerate(reversed(rho_array)):
        flx, car = _gen_fluxsurface_mesh(obj, rho**2, **kwargs)

        # Triangulation only needs to be done once since all the meshes
        # are generated using the same point ordering.
        if ii == 0:
            tri = scipy.spatial.Delaunay(flx[:, 1:])

        faces = tri.simplices

        color = np.array(cm.to_rgba(rho))*255
        color = color.astype(int)
        color_string = f'rgb({color[0]}, {color[1]}, {color[2]})'

        trace = go.Mesh3d(
            x=car[:, 0],
            y=car[:, 1],
            z=car[:, 2],
            i=faces[:, 0],
            j=faces[:, 1],
            k=faces[:, 2],
            color=color_string,
            flatshading=flatshading,
            opacity=alpha
        )

        figure.add_trace(trace)


def _make_plotly_color(color, alpha=None):
    if isinstance(color, str):
        color_str = color
    else:
        raise Exception('Only plotly color strings are currently supported.')

    return color_str


# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir pablant <npablant@pppl.gov>

Some utilities to help with version inspection.
"""

from xicsrt import _version

from packaging import version
from xicsrt.util import mirlogging
log = mirlogging.getLogger('xicsrt')

def warn_version(v_string):
    v_input = version.parse(v_string)
    v_current = version.parse(_version.__version__)

    if (v_input.major < v_current.major) or ((v_input.major == v_current.major) and (v_input.minor < v_current.minor)):
        log.warning('This config is for an older version of xicsrt. Some options may have changed.')
    elif (v_input.major > v_current.major) or ((v_input.major == v_current.major) and (v_input.minor > v_current.minor)):
        log.warning('This config is for a newer version of xicsrt. Please upgrade xicsrt (pip install --upgrade xicsrt).')



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>

A collections of routines to simplify interactive use of XICSRT.
"""
import numpy as np

from xicsrt import xicsrt_config
from xicsrt.objects._Dispatcher import Dispatcher

def get_element(config_user, name, section=None, initialize=True):
    """
    Retrieves an raytracing element (source, optic or filter) object.
    """
    config = xicsrt_config.get_config(config_user)
    if section is None:
        section = _find_element_section(config, name)

    disp = Dispatcher(config, section)
    disp.instantiate(name)
    obj = disp.get_object(name)
    if initialize:
        obj.setup()
        obj.check_param()
        obj.initialize()
    return obj

def _find_element_section(config, name):
    """
    Search config for the given element name and return the section.
    """
    section_list = ['optics', 'sources', 'filters']
    out_list = []
    for section in section_list:
        if name in config[section]:
            out_list.append(section)

    if len(out_list) == 0:
        raise Exception(f'Could not find element: {name} in any section.')
    if len(out_list) > 1:
        raise Warning(f'Element name: {name} was found in more than one section.'
                      f' Please provide an explicit section name.')

    return out_list[0]



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`InteractNone` class.
"""

import numpy as np
from copy import deepcopy

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractObject import InteractObject

@dochelper
class InteractNone(InteractObject):
    """
    No interaction with surface, rays will pass through unchanged.
    """

    # The behavior is identical to that of the base class.
    pass


# -*- coding: utf-8 -*-
"""
.. Authors:
   Conor Perks <cjperks@psfc.mit.edu>
   Novimir Pablant <npablant@pppl.gov>

Define the :class:`ShapeCylinder` class.
"""
import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.tools import xicsrt_math as xm
from xicsrt.optics._ShapeObject import ShapeObject

@dochelper
class ShapeCylinder(ShapeObject):
    """
    A cylindrical shape.
    This class defines intersections with a cylinder.
    """

    def default_config(self):
        """
        radius : float (1.0)
          The radius of the cylinder.

        convex : bool (False)
          If True the optic will have a convex curvature, if False the surface
          will have a concave curvature (the default).
        """
        config = super().default_config()
        config['radius'] = 1.0
        config['convex'] = False
        return config

    def initialize(self):
        super().initialize()
        # Finds center location of the cylinder geometry (in global coordinates).
        if self.param['convex']:
            sign = -1
        else:
            sign = 1
        self.param['center'] = sign*self.param['radius'] * self.param['zaxis'] + self.param['origin']

    def intersect(self, rays):
        dist, mask = self.intersect_distance(rays)
        xloc = self.location_from_distance(rays, dist, mask)
        norm = self.intersect_normal(xloc, mask)

        return xloc, norm, mask

    def intersect_distance(self, rays):
        """
        Calulate the distance to the intersection of the rays with the
        cylindrical optic.

        This calculation is copied from:
        https://mrl.cs.nyu.edu/~dzorin/rend05/lecture2.pdf
        """

        # Create short variable names for readability.
        O = rays['origin']
        D = rays['direction']
        m = rays['mask']

        # Initializes arrays to store intersection variables
        distance = np.full(m.shape, np.nan, dtype=np.float64)

        # Create short variable names for readability.
        pa = self.param['center']
        va = self.param['xaxis']
        r = self.param['radius']

        # Calculate quadratic formula coefficients for t
        #
        # The cylinter axis parameterized as pa + t_axis*va
        # Below einsum is used to calculate dot products.

        # Length between ray origin and cylinder center
        dp = O-pa
        dot_Dva = np.einsum('ij,j->i', D, va)
        dot_dpva = np.einsum('ij,j->i', dp, va)
        A1 = D - dot_Dva[:,None]*va[None,:]
        B1 = dp - dot_dpva[:,None]*va[None,:]

        # Coefficients for f(t) = A*t**2 + B*t + C = 0
        A = np.einsum('ij,ij->i', A1, A1)
        B = 2*np.einsum('ij,ij->i', A1, B1)
        C = np.einsum('ij,ij->i', B1, B1)-r**2

        # Solves the quadratic formula for t
        dis = B**2-4*A*C

        # Update the mask to exclude rays that don't hit the cylinder.
        m[m] &= (dis[m] >= 0)

        t_0      = np.empty(m.shape, dtype=np.float64)
        t_1      = np.empty(m.shape, dtype=np.float64)
        t_0[m] = (-B[m] - np.sqrt(dis[m]))/(2*A[m])
        t_1[m] = (-B[m] + np.sqrt(dis[m]))/(2*A[m])

        # Distance traveled by the ray before hitting the optic surface with
        # the chosen curvature.
        if self.param['convex']:
            distance[m] = np.where(t_0[m] < t_1[m], t_0[m], t_1[m])
        else:
            distance[m] = np.where(t_0[m] > t_1[m], t_0[m], t_1[m])

        return distance, m

    def intersect_normal(self, xloc, mask):
        """
        Calculates the normal vector from the intersection with the cylinder surface
        """

        # Create short variable names for readability.
        m = mask
        pa = self.param['center']
        va = self.param['xaxis']

        # Initializes quantities
        norm = np.full(xloc.shape, np.nan, dtype=np.float64)
        pa_proj = np.full(xloc.shape, np.nan, dtype=np.float64)

        # A normal vector on a cylinder is orthogonal to the cylinder axis
        # As such, we need to project the 'center' of the cylinder onto the same
        # plane as the normal vector
        dummy = np.einsum('ij,j->i', pa-xloc[m], va)
        pa_proj[m] = pa - dummy[:,None]*va[None,:]

        # Calculates the normal vector from intersection point to cylinder center
        norm[m] = xm.normalize(pa_proj[m] - xloc[m])
        return norm
